"""Import required create_report."""

from .report import build_bucket_table, psi, iv
from .plotting import plot_bucket_table, weight_plot

__all__ = ["build_bucket_table", "plot_bucket_table", "psi", "iv", "weight_plot"]



import ast
import json

import pandas as pd

from skorecard.apps.app_utils import determine_boundaries, is_increasing, is_sequential
from skorecard.reporting import build_bucket_table
from skorecard.utils.exceptions import NotInstalledError

# Dash + dependencies
try:
    from dash import no_update
    from dash.dependencies import Input, Output, State
except ModuleNotFoundError:
    Input = NotInstalledError("dash", "dashboard")
    Output = NotInstalledError("dash", "dashboard")
    State = NotInstalledError("dash", "dashboard")
    dash_table = NotInstalledError("dash_table", "dashboard")


def add_bucketing_callbacks(self, X, y):
    """
    Adds callbacks to the interactive bucketing app.

    Meant for normal bucketers, not two step BucketingProcess.
    """
    app = self.app
    add_common_callbacks(self)

    @app.callback(
        [Output("input_map", "value")],
        [
            Input("input_column", "value"),
        ],
    )
    def update_input_map(col):
        """Update bucketer map."""
        input_map = self.features_bucket_mapping_.get(col).map
        col_type = self.features_bucket_mapping_.get(col).type

        if col_type == "categorical":
            # We also allow for treating numerical as categoricals
            # if key is a string, we'll need to quote them
            if isinstance(list(input_map.keys())[0], str):
                str_repr = ",\n\t".join([f"'{k}': {v}" for k, v in input_map.items()])
            else:
                str_repr = ",\n\t".join([f"{k}: {v}" for k, v in input_map.items()])
            str_repr = f"{{\n\t{str_repr}\n}}"
        else:
            str_repr = str(input_map)
        return [str_repr]

    @app.callback(
        [Output("input_map_helptext", "children")],
        [
            Input("input_column", "value"),
        ],
    )
    def update_input_map_feedback(col):
        col_type = self.features_bucket_mapping_.get(col).type
        right = self.features_bucket_mapping_.get(col).right
        if col_type == "categorical":
            msg = "Edit the prebucket mapping dictionary, e.g. {'value' : 'pre-bucket'}"
        if col_type == "numerical" and right:
            msg = "Edit the prebucket mapping boundaries. "
            msg += "Values up to and including the boundary are put into a bucket (right=True)"
        if col_type == "numerical" and not right:
            msg = "Edit the prebucket mapping boundaries. "
            msg += "Values up to but not including the boundary are put into a bucket (right=False)"
        return [msg]

    @app.callback(
        [
            Output("bucket_table", "data"),
            Output("graph-bucket", "figure"),
            Output("input_map", "invalid"),
            Output("input_map_feedback", "children"),
        ],
        [Input("input_map", "value")],
        [State("input_column", "value")],
    )
    def get_bucket_table(input_map, col):
        """Loads the table and the figure, when the input_map changes."""
        col_type = self.features_bucket_mapping_.get(col).type

        # Load the object from text input into python object
        if col_type == "numerical":
            try:
                input_map = json.loads(input_map)
                assert len(input_map) > 0
            except Exception:
                msg = "Make sure the input is properly formatted as a list"
                return no_update, no_update, True, [msg]
            # validate input
            if not is_increasing(input_map):
                return no_update, no_update, True, ["Make sure the list values are in increasing order"]
        else:
            try:
                # note using ast.literal_eval is not safe
                # for use when you don't trust the user input
                # in this case, it's a local user using his/her own kernel
                # note: we're using literal_eval because JSON enforces quoted keys
                input_map = ast.literal_eval(input_map)
                # re-sort on value, key
                input_map = dict(sorted(input_map.items(), key=lambda x: (x[1], x[0])))
            except Exception:
                msg = "Make sure the input is properly formatted as a dictionary"
                return no_update, no_update, True, [msg]
            # validate input
            if not min(input_map.values()) == 0:
                msg = "Dictionary values (buckets) must start at 0"
                return no_update, no_update, True, [msg]
            if not is_sequential(list(input_map.values())):
                msg = "Dictionary values (buckets) must be sequentially increasing with steps of 1"
                return no_update, no_update, True, [msg]

        # Update the fit for this specific column
        special = self.features_bucket_mapping_.get(col).specials
        right = self.features_bucket_mapping_.get(col).right
        # Note we passed X, y to add_bucketing_callbacks() so they are available here.
        # make sure to re-generate the summary table
        self._update_column_fit(
            X=X, y=y, feature=col, special=special, splits=input_map, right=right, generate_summary=True
        )

        # Retrieve the new bucket tables and plots
        table = self.bucket_table(col)
        # unsupervised bucketers don't have an event rate.
        if "Event Rate" in table.columns:
            table["Event Rate"] = round(table["Event Rate"] * 100, 2)
        fig = self.plot_bucket(col)
        # remove title from plot
        fig.update_layout(title="")
        return table.to_dict("records"), fig, False, no_update


def add_bucketing_process_callbacks(self, X, y):
    """
    Adds callbacks to the interactive bucketing app.

    Meant two step BucketingProcess.
    """
    app = self.app
    add_common_callbacks(self)

    @app.callback(
        [Output("bucketingprocess-helptext", "style")],
        [
            Input("input_column", "value"),
        ],
    )
    def update_sidebar_helptext(col):
        return [{"display": "block"}]

    @app.callback(
        [Output("input_map", "value")],
        [
            Input("input_column", "value"),
        ],
    )
    def update_input_map(col):
        """Update bucketer map."""
        input_map = self.pre_pipeline_.features_bucket_mapping_.get(col).map
        col_type = self.pre_pipeline_.features_bucket_mapping_.get(col).type

        if col_type == "categorical":
            # We also allow for treating numerical as categoricals
            # if key is a string, we'll need to quote them
            if isinstance(list(input_map.keys())[0], str):
                str_repr = ",\n\t".join([f"'{k}': {v}" for k, v in input_map.items()])
            else:
                str_repr = ",\n\t".join([f"{k}: {v}" for k, v in input_map.items()])
            str_repr = f"{{\n\t{str_repr}\n}}"
        else:
            str_repr = str(input_map)
        return [str_repr]

    @app.callback(
        [Output("input_map_helptext", "children")],
        [
            Input("input_column", "value"),
        ],
    )
    def update_input_map_feedback(col):
        col_type = self.pre_pipeline_.features_bucket_mapping_.get(col).type
        right = self.pre_pipeline_.features_bucket_mapping_.get(col).right
        if col_type == "categorical":
            msg = "Edit the prebucket mapping dictionary, e.g. {'value' : 'pre-bucket'}"
        if col_type == "numerical" and right:
            msg = "Edit the prebucket mapping boundaries. "
            msg += "Values up to and including the boundary are put into a bucket (right=True)"
        if col_type == "numerical" and not right:
            msg = "Edit the prebucket mapping boundaries. "
            msg += "Values up to but not including the boundary are put into a bucket (right=False)"
        return [msg]

    @app.callback(
        [
            Output("pre_bucket_table", "data"),
            Output("input_map", "invalid"),
            Output("input_map_feedback", "children"),
        ],
        [Input("input_map", "value")],
        [State("input_column", "value")],
    )
    def get_prebucket_table(input_map, col):
        """Loads the table and the figure, when the input_map changes."""
        col_type = self.pre_pipeline_.features_bucket_mapping_.get(col).type

        # Load the object from text input into python object
        if col_type == "numerical":
            try:
                input_map = json.loads(input_map)
                assert len(input_map) > 0
            except Exception:
                msg = "Make sure the input is properly formatted as a list"
                return no_update, True, [msg]
            # validate input
            if not is_increasing(input_map):
                return no_update, True, ["Make sure the list values are in increasing order"]
        else:
            try:
                # note using ast.literal_eval is not safe
                # for use when you don't trust the user input
                # in this case, it's a local user using his/her own kernel
                input_map = ast.literal_eval(input_map)
                # re-sort on value, key
                input_map = dict(sorted(input_map.items(), key=lambda x: (x[1], x[0])))
            except Exception:
                msg = "Make sure the input is properly formatted as a dictionary"
                return no_update, True, [msg]
            # validate input
            if not min(input_map.values()) == 0:
                msg = "Dictionary values (buckets) must start at 0"
                return no_update, True, [msg]
            if not is_sequential(list(input_map.values())):
                msg = "Dictionary values (buckets) must be sequentially increasing with steps of 1"
                return no_update, True, [msg]

        # Update the fit for this specific column
        # Note we passed X, y to add_bucketing_process_callbacks() so they are available here.
        # make sure to re-generate the summary table
        for step in self.pre_pipeline_.steps:
            if col in step[1].variables:
                step[1]._update_column_fit(
                    X=X,
                    y=y,
                    feature=col,
                    special=self._prebucketing_specials.get(col, {}),
                    splits=input_map,
                    right=self.pre_pipeline_.features_bucket_mapping_.get(col).right,
                    generate_summary=True,
                )

        self.prebucket_tables_[col] = build_bucket_table(
            X, y, column=col, bucket_mapping=self.pre_pipeline_.features_bucket_mapping_.get(col)
        )
        # Re-calculate the BucketingProcess summary
        self._generate_summary(X, y)

        # Retrieve the new bucket tables and plots
        table = self.prebucket_table(col)
        # unsupervised bucketers don't have an event rate.
        if "Event Rate" in table.columns:
            table["Event Rate"] = round(table["Event Rate"] * 100, 2)
        return table.to_dict("records"), False, no_update

    @app.callback(
        [
            Output("bucket_table", "data"),
            Output("graph-prebucket", "figure"),
            Output("graph-bucket", "figure"),
            Output("bucket-error-msg", "children"),
            Output("bucket-error-msg", "style"),
        ],
        [Input("pre_bucket_table", "data")],
        [State("input_column", "value")],
    )
    def get_bucket_table(prebucket_table, col):
        # Get the input from the prebucket table
        new_buckets = pd.DataFrame()
        new_buckets["pre_buckets"] = [row.get("pre-bucket") for row in prebucket_table]
        input_buckets = [int(row.get("bucket")) for row in prebucket_table]
        new_buckets["buckets"] = input_buckets

        # Input validation
        bucket_mapping = self.pipeline_.features_bucket_mapping_.get(col)
        if not is_sequential(sorted(input_buckets)):
            msg = "Buckets must start at 0 and increase by 1"
            return no_update, no_update, no_update, [msg], {"display": "block"}

        # Determine the boundaries from the buckets set in the prebucket table
        boundaries = determine_boundaries(new_buckets, bucket_mapping)

        # Update the fit for this specific column
        # Note we passed X, y to add_bucketing_process_callbacks() so they are available here.
        # make sure to re-generate the summary table
        for step in self.pipeline_.steps:
            if col in step[1].variables:
                step[1]._update_column_fit(
                    X=X,
                    y=y,
                    feature=col,
                    special=bucket_mapping.specials,
                    splits=boundaries,
                    right=bucket_mapping.right,
                    generate_summary=True,
                )

        # update the BucketingProcess bucket tables.
        X_prebucketed = self.pre_pipeline_.transform(X)
        self.bucket_tables_[col] = build_bucket_table(
            X_prebucketed,
            y,
            column=col,
            bucket_mapping=self.pipeline_.features_bucket_mapping_.get(col),
        )

        # Re-calculate the BucketingProcess summary
        self._generate_summary(X, y)

        # Get the updated tables and figures
        table = self.bucket_table(col)
        # unsupervised bucketers don't have an event rate.
        if "Event Rate" in table.columns:
            table["Event Rate"] = round(table["Event Rate"] * 100, 2)
        prebucket_fig = self.plot_prebucket(col)
        prebucket_fig.update_layout(title="")
        fig = self.plot_bucket(col)
        fig.update_layout(title="")

        return table.to_dict("records"), prebucket_fig, fig, no_update, {"display": "none"}


def add_common_callbacks(self):
    """
    Add dash callbacks.

    Common callbacks for the normal bucketer app and the BucketingProcess app.
    """
    app = self.app

    @app.callback(
        [
            Output("column_title", "children"),
            Output("column_type", "children"),
        ],
        [
            Input("input_column", "value"),
        ],
    )
    def update_column_title(col):
        """Update the content title."""
        col_type = self.features_bucket_mapping_.get(col).type
        return [f"Feature '{col}'"], [col_type]

    @app.callback(
        [Output("code_export", "content")],
        [Input("input_map", "value")],
    )
    def update_code_export(input_map):
        return [f"UserInputBucketer({self.features_bucket_mapping_.as_dict()})"]



import pandas as pd
import pytest

from skorecard.bucketers import AsIsNumericalBucketer
from skorecard.utils.exceptions import NotPreBucketedError


def test_manual_transformation(df):
    """
    Test that we can use an example dict for ManualBucketTransformer.
    """
    ref = pd.DataFrame({"a": [1, 3, 5]})
    ref2 = pd.DataFrame({"a": [1.001, 2.999, 3, 6]})

    ai = AsIsNumericalBucketer()
    ai.fit(ref)
    assert ai.bucket_table("a")["label"][1] == "(-inf, 1.0]"
    assert list(ai.transform(ref2)["a"].values) == [1, 1, 1, 3]

    ai = AsIsNumericalBucketer(right=False)
    ai.fit(ref)

    assert ai.bucket_table("a")["label"][1] == "[-inf, 1.0)"
    assert list(ai.transform(ref2)["a"].values) == [1, 1, 2, 3]


def test_fit_transform(df):
    """
    Test basic transform.
    """
    features = ["LIMIT_BAL", "EDUCATION", "MARRIAGE"]  # BILL_AMT1 has > 100 unique values
    X = df[features]
    y = df["default"].values

    ai = AsIsNumericalBucketer()
    ai.fit_transform(X, y)

    with pytest.raises(NotPreBucketedError):
        X = df[["BILL_AMT1"]]
        ai = AsIsNumericalBucketer()
        ai.fit_transform(X, y)



import numpy as np
import pandas as pd

from skorecard import Skorecard
from skorecard.datasets import load_uci_credit_card


def test_bucket_table_woe_values():
    """Checks whether or not the WoE-values of bucket_table()are equivalent to those in the transformed data."""
    data = load_uci_credit_card()
    X = data["data"]
    y = data["target"]

    model = Skorecard()
    model = model.fit(X, y)
    X_woe = model.woe_transform(X)
    for c in X.columns:
        bucket_table = model.bucket_table(c)
        b_tab_woes = set(bucket_table["WoE"])
        b_tab_woes = {x for x in b_tab_woes if pd.notna(x)}
        data_woes = set(np.round(X_woe[c].value_counts().index, 3))
        assert b_tab_woes == data_woes
        iv_lt_zero = [x < 0 for x in bucket_table["IV"]]
        assert True not in iv_lt_zero



import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from skorecard.utils.exceptions import NotInstalledError

try:
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
except ModuleNotFoundError:
    NotInstalledError("plotly", "reporting")

try:
    from IPython.core.display import Image
except ModuleNotFoundError:
    NotInstalledError("psutil")


def make_plot_figure(bucket_table: pd.DataFrame, line="event_rate"):
    """
    Make a plotly object out of a table. The line defines what is plotted on the y-axis.
    """
    # To support both pre-buckets and buckets
    if "pre-bucket" in bucket_table.columns:
        buckets = [b for b in bucket_table["pre-bucket"].values]
    else:
        buckets = [b for b in bucket_table["bucket"].values]

    plotdf = pd.DataFrame(
        {
            "bucket": buckets,
            "label": [x for x in bucket_table["label"].values],
            "counts": [int(count) for count in bucket_table["Count"].values],
            "counts %": [float(count / 100) for count in bucket_table["Count (%)"].values],
        }
    )

    if line == "event_rate":
        column_to_plot = "Event Rate"
    elif line == "woe":
        column_to_plot = "WoE"
    else:
        raise AttributeError(f"line {line} must be event_rate or woe")

    # If the bucket_table is built without any 'y' information
    # we won't know any event rate rates either
    # and thus we need to output a simpler plot
    if column_to_plot in bucket_table.columns:
        plotdf[column_to_plot] = [event for event in bucket_table[column_to_plot].values]
        fig = make_subplots(specs=[[{"secondary_y": True}]])

        # Add trace with event rate
        fig.add_trace(
            go.Scatter(
                x=plotdf["label"],
                y=plotdf[column_to_plot],
                name=column_to_plot,
                line=dict(color="#454c57"),
            ),
            secondary_y=True,
        )
        fig.update_yaxes(title_text=f"Bucket {column_to_plot}", secondary_y=True, tickformat=",.0%")

    else:
        fig = make_subplots(specs=[[{"secondary_y": False}]])

    # Add trace with counts
    fig.add_trace(
        go.Bar(x=plotdf["label"], y=plotdf["counts %"], name="Bucket count percentage"),
        secondary_y=False,
    )
    fig.update_yaxes(title_text="bucket size", secondary_y=False, tickformat=",.0%")

    # If we have bucket information, use that to colorize the bars
    # that means a prebucket table without information on the corresponding buckets
    # won't have bars colorized.
    if "bucket" in bucket_table.columns:
        bar_colors = [get_bucket_color(i) for i in bucket_table["bucket"].values]
        fig.update_traces(marker=dict(color=bar_colors), selector=dict(type="bar"))

    # Other stuff
    fig.update_layout(transition_duration=50)
    fig.update_layout(showlegend=False)
    fig.update_xaxes(type="category")
    fig.update_layout(
        margin=dict(l=20, r=20, t=40, b=20),
        height=350,
    )
    return fig


def plot_prebucket_table(
    prebucket_table,
    column="",
    line="",
    format=None,
    scale=None,
    width=None,
    height=None,
):
    """
    Given the prebucketed data, plot the pre-buckets.

    Args:
        prebucket_table (pd.DataFrame): the table of the prebucketed data
        column (str): The column to plot
        line (str): The line to plot on the secondary axis
        format (str): The format of the image, e.g. 'png'. The default returns a plotly fig
        scale: If format is specified, the scale of the image
        width: If format is specified, the width of the image
        height: If format is specified, the image of the image

    Returns:
        fig of desired format
    """
    fig = make_plot_figure(prebucket_table, line)

    fig.update_layout(title=f"pre-buckets: {column}".strip())
    fig.update_layout(xaxis_title=f"{column} pre-buckets".strip())

    if format is not None:
        img_bytes = fig.to_image(format=format, scale=scale, width=width, height=height)
        fig = Image(img_bytes)
    return fig


def plot_bucket_table(bucket_table, column="", line="", format=None, scale=None, width=None, height=None):
    """
    Given the bucketed data, plot the buckets with Event Rate.

    Args:
        bucket_table (pd.DataFrame): the table of the bucketed data
        format (str): The format of the image, e.g. 'png'. The default returns a plotly fig
        line (str): The line to plot on the secondary axis
        scale: If format is specified, the scale of the image
        width: If format is specified, the width of the image
        height: If format is specified, the image of the image

    Returns:
        plotly fig
    """
    fig = make_plot_figure(bucket_table, line)

    fig.update_layout(title=f"buckets: {column}".strip())
    fig.update_layout(xaxis_title=f"{column} buckets".strip())

    if format is not None:
        img_bytes = fig.to_image(format=format, scale=scale, width=width, height=height)
        fig = Image(img_bytes)

    return fig


def get_bucket_color(i):
    """Return diverging color for unique buckets.

    Generated using:

    ```python
    import seaborn as sns
    colors = sns.color_palette("Set2")
    rgbs = []
    for r,g,b in list(colors):
        rgbs.append(
            f"rgb({int(r*255)},{int(g*255)},{int(b*255)})"
        )
    ```

    Args:
        i: number to return a color string for
    """
    if i < 0:
        return "rgb(189,189,189)"  # light grey for specials
    elif i % 8 == 0:
        return "rgb(102,194,165)"
    elif i % 8 == 1:
        return "rgb(252,141,98)"
    elif i % 8 == 2:
        return "rgb(141,160,203)"
    elif i % 8 == 3:
        return "rgb(231,138,195)"
    elif i % 8 == 4:
        return "rgb(166,216,84)"
    elif i % 8 == 5:
        return "rgb(255,217,47)"
    elif i % 8 == 6:
        return "rgb(229,196,148)"
    elif i % 8 == 7:
        return "rgb(179,179,179)"
    else:
        raise NotImplementedError("")


class PlotPreBucketMethod:
    """
    Add methods for plotting bucketing tables to another class.

    To be used with skorecard.pipeline.BucketingProcess and skorecard.bucketers.BaseBucketer
    """

    def plot_prebucket(
        self,
        column,
        line="event_rate",
        format=None,
        scale=None,
        width=None,
        height=None,
    ):
        """
        Generates the prebucket table and produces a corresponding plotly plot.

        Args:
            column: The column we want to visualise
            line: The line to plot on the secondary axis. Default is Event Rate.
            format: The format of the image, such as 'png'. The default None returns a plotly image.
            scale: If format is specified, the scale of the image
            width: If format is specified, the width of the image
            height: If format is specified, the image of the image

        Returns:
            plot: plotly fig
        """
        assert line in ["event_rate", "woe"], "line must be event_rate or woe"
        if isinstance(self, Pipeline):
            check_is_fitted(self.steps[0][1])
        else:
            check_is_fitted(self)
        return plot_prebucket_table(
            prebucket_table=self.prebucket_table(column),
            line=line,
            column=column,
            format=format,
            scale=scale,
            width=width,
            height=height,
        )


class PlotBucketMethod:
    """
    Add methods for plotting bucketing tables to another class.

    To be used with skorecard.pipeline.BucketingProcess and skorecard.bucketers.BaseBucketer
    """

    def plot_bucket(
        self,
        column,
        line="event_rate",
        format=None,
        scale=None,
        width=None,
        height=None,
    ):
        """
        Plot the buckets.

        Args:
            column: The column we want to visualise
            line: The line to plot on the secondary axis. Default is Event Rate.
            format: The format of the image, such as 'png'. The default None returns a plotly image.
            scale: If format is specified, the scale of the image
            width: If format is specified, the width of the image
            height: If format is specified, the image of the image

        Returns:
            plot: plotly fig
        """
        assert line in ["event_rate", "woe"], "line must be event_rate or woe"
        if isinstance(self, Pipeline):
            check_is_fitted(self.steps[0][1])
        else:
            check_is_fitted(self)

        return plot_bucket_table(
            bucket_table=self.bucket_table(column=column),
            line=line,
            column=column,
            format=format,
            scale=scale,
            width=width,
            height=height,
        )


def weight_plot(stats: pd.DataFrame, format=None, scale=None, width=None, height=None):
    """
    Generates a weight plot(plotly chart) from `stats`
    Example:

    ```from skorecard.datasets import load_uci_credit_card
    from skorecard.bucketers import EqualFrequencyBucketer
    from skorecard.linear_model import LogisticRegression
    from skorecard.reporting import weight_plot
    from sklearn.pipeline import Pipeline
    X, y = load_uci_credit_card(return_X_y=True)
    pipeline = Pipeline([
        ('bucketer', EqualFrequencyBucketer(n_bins=10)),
        ('clf', LogisticRegression(calculate_stats=True))
    ])
    pipeline.fit(X, y)
    stats = pipeline.named_steps['clf'].get_stats()

    weight_plot(stats)```

    Args:
        stats: The statistics to display
        format: The format of the image, such as 'png'. The default None returns a plotly image.
        scale: If format is specified, the scale of the image
        width: If format is specified, the width of the image
        height: If format is specified, the image of the image
    """
    fig = go.Figure()

    stats["conf_interval_0.025"] = stats["Coef."] - 1.96 * stats["Std.Err"]
    stats["conf_interval_0.975"] = stats["Coef."] + 1.96 * stats["Std.Err"]

    fig.add_trace(
        go.Scatter(
            x=stats["Coef."],
            y=stats["Coef."].index,
            line=dict(color="#42C4F7", width=2),
            mode="markers",
            error_x=dict(
                type="data",
                symmetric=False,
                array=stats["conf_interval_0.975"] - stats["Coef."],
                arrayminus=stats["Coef."] - stats["conf_interval_0.025"],
                color="#68BBE3",
            ),
        )
    )

    fig.add_shape(type="line", x0=0, y0=0, x1=0, y1=len(stats), line=dict(color="#3f3f3f", width=3, dash="dash"))

    fig.update_layout(
        title="Regression Meta Analysis - Weight Plot",
        xaxis_title="Weight Estimates",
        yaxis_title="Variable",
        xaxis_showgrid=False,
        yaxis_showgrid=False,
    )
    fig.update_layout(template="simple_white")

    if format is not None:
        img_bytes = fig.to_image(format=format, scale=scale, width=width, height=height)
        fig = Image(img_bytes)
    return fig

    return fig



import io
import pkgutil

import pandas as pd
from sklearn.datasets import fetch_openml


def load_uci_credit_card(return_X_y=False, as_frame=False):
    """Loads the UCI Credit Card Dataset.

    This dataset contains a sample of [Default of Credit Card Clients Dataset](https://www.kaggle.com/uciml/default-of-credit-card-clients-dataset).

    Example:

    ```python
    from skorecard import datasets
    df = datasets.load_uci_credit_card(as_frame=True)
    ```

    Args:
        return_X_y (bool): If True, returns `(data, target)` instead of a dict object.
        as_frame (bool): give the pandas dataframe instead of X, y matrices (default=False).

    Returns: (pd.DataFrame, dict or tuple) features and target, with as follows:
        - if as_frame is True: returns pd.DataFrame with y as a target
        - return_X_y is True: returns a tuple: (X,y)
        - is both are false (default setting): returns a dictionary where the key `data` contains the features,
        and the key `target` is the target

    """  # noqa
    file = pkgutil.get_data("skorecard", "data/UCI_Credit_Card.zip")
    df = pd.read_csv(io.BytesIO(file), compression="zip")
    df = df.rename(columns={"default.payment.next.month": "default"})
    if as_frame:
        return df[["EDUCATION", "MARRIAGE", "LIMIT_BAL", "BILL_AMT1", "default"]]
    X, y = (
        df[["EDUCATION", "MARRIAGE", "LIMIT_BAL", "BILL_AMT1"]],
        df["default"].values,
    )
    if return_X_y:
        return X, y

    return {"data": X, "target": y}


def load_credit_card(return_X_y=False, as_frame=False):
    """
    Loads the complete UCI Credit Card Dataset, by fetching it from open_ml.

    Args:
        return_X_y:  (bool) If True, returns ``(data, target)`` instead of a dict object.
        as_frame: (bool) give the pandas dataframe instead of X, y matrices (default=False).

    Returns: (pd.DataFrame, dict or tuple) features and target, with as follows:
        - if as_frame is True: returns pd.DataFrame with y as a target
        - return_X_y is True: returns a tuple: (X,y)
        - is both are false (default setting): returns a dictionary where the key `data` contains the features,
        and the key `target` is the target

    """
    try:
        data = fetch_openml(
            name="default-of-credit-card-clients",
            version=1,
            data_home=None,
            cache=True,
            as_frame=as_frame,
            return_X_y=return_X_y,
            parser="liac-arff",
        )
    except Exception as e:
        # update the error message with a more helpful message.
        error_msg = (
            "Cannot retrieve the dataset from repository. Make sure there is no firewall blocking "
            "the connection.\nAlternatively, download it manually from https://www.openml.org/d/42477"
        )
        raise type(e)(f"{e.args[0]}\n{error_msg}")

    # The target is by default encoded as a string.
    # Ensure it is returned as a integer.
    if as_frame:
        data = data["frame"]
        data["y"] = data["y"].astype(int)
    if return_X_y:
        X = data[0]
        y = data[1]
        y = y.astype(int)
        return X, y

    return data



import numpy as np
import pytest
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.utils.validation import check_is_fitted

from skorecard import Skorecard
from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer, OrdinalCategoricalBucketer
from skorecard.features_bucket_mapping import FeaturesBucketMapping
from skorecard.pipeline import BucketingProcess
from skorecard.utils import BucketerTypeError


def run_checks(X, y, bucketer, features, expected_probas):
    """
    Run some standard asserts on Skorecard instance.
    """
    skorecard_model = Skorecard(bucketing=bucketer, variables=features, calculate_stats=True)
    skorecard_model.fit(X, y)

    # make sure sklearn recognizes this as fitted
    check_is_fitted(skorecard_model)
    check_is_fitted(skorecard_model.pipeline_.steps[-1][1])
    if isinstance(skorecard_model.bucketing_, Pipeline):
        check_is_fitted(skorecard_model.bucketing_.steps[0][1])
    else:
        check_is_fitted(skorecard_model.bucketing_)

    # Make sure proper attributes are there
    assert isinstance(skorecard_model.bucketing_.features_bucket_mapping_, FeaturesBucketMapping)
    assert isinstance(skorecard_model.bucketing_.bucket_tables_, dict)

    top_probas = skorecard_model.predict_proba(X)[:2]
    np.testing.assert_array_almost_equal(top_probas, expected_probas, decimal=2)

    # assure naming convention is fixed and ordered
    assert [names for names in skorecard_model.pipeline_.named_steps] == [
        "bucketer",
        "encoder",
        "column_selector",
        "model",
    ]

    if len(features) == 0:
        features = X.columns.tolist()

    # test bucketers did not change
    if bucketer is not None:
        assert (
            skorecard_model.bucket_transform(X)[features].head(3).equals(bucketer.fit_transform(X, y)[features].head(3))
        )

    # check that columns are properly selected
    assert skorecard_model.bucket_transform(X).shape == X.shape
    assert skorecard_model.woe_transform(X).shape == X.shape
    # Last transformer selects the features
    assert skorecard_model.pipeline_[:-1].transform(X).shape == (X.shape[0], len(features))

    # test that the stats showcase only the selected features
    assert skorecard_model.get_stats().index.tolist() == ["const"] + features

    # Test bucket table works
    assert (
        skorecard_model.bucket_table("LIMIT_BAL").shape[0]
        == skorecard_model.pipeline_[:1].transform(X)["LIMIT_BAL"].nunique() + 1
    )

    # prebucketing methods
    if not isinstance(skorecard_model.bucketing_, BucketingProcess):
        # test a BucketerTypeError is raised if the plot_prebucket('LIMIT_BAL') function is called (not defined
        # #for standard bucketers)
        with pytest.raises(BucketerTypeError):
            skorecard_model.plot_prebucket("LIMIT_BAL")

        # test a BucketerTypeError is raised if the prebucket_table function is called
        #  (not defined for standard bucketers)
        with pytest.raises(BucketerTypeError):
            skorecard_model.prebucket_table("LIMIT_BAL")
    else:
        prebucket_table = skorecard_model.prebucket_table("LIMIT_BAL")
        bucket_table = skorecard_model.bucket_table("LIMIT_BAL")
        assert prebucket_table.shape[0] >= bucket_table.shape[0]


def test_skorecard_with_num_bucketers(df):
    """Test a workflow, with numerical bucketer and numerical features alone."""
    X = df.drop("default", axis=1)
    y = df["default"]

    features = ["LIMIT_BAL", "BILL_AMT1"]
    bucketer = DecisionTreeBucketer(variables=features, max_n_bins=5)
    expected_probas = np.array([[0.841, 0.159], [0.738, 0.262]])

    run_checks(X, y, bucketer, features, expected_probas)

    # and with features not defined
    features = []
    expected_probas = np.array([[0.88, 0.12], [0.73, 0.27]])
    run_checks(X, y, bucketer, features, expected_probas)


def test_passing_kwargs(df):
    """Test passing keyword args to LR."""
    X = df.drop("default", axis=1)
    y = df["default"]

    features = ["LIMIT_BAL", "BILL_AMT1"]
    bucketer = DecisionTreeBucketer(variables=features, max_n_bins=5)

    skorecard_model = Skorecard(
        bucketing=bucketer,
        variables=features,
        lr_kwargs={"penalty": None, "C": 1, "multi_class": "ovr", "n_jobs": 1, "max_iter": int(1e3)},
    )
    skorecard_model.fit(X, y)

    assert skorecard_model.bucket_transform(X).shape == X.shape
    assert skorecard_model.woe_transform(X).shape == X.shape


def test_skorecard_with_bucketing_process(df):
    """Test a workflow, with Bucketin Process."""
    X = df.drop("default", axis=1)
    y = df["default"]

    features = ["LIMIT_BAL", "BILL_AMT1"]
    cat_features = [col for col in X.columns if col not in features]

    prebucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=features, max_n_bins=100),
        OrdinalCategoricalBucketer(variables=cat_features, tol=0.01),
    )
    bucketing_pipeline = make_pipeline(
        OptimalBucketer(variables=features, max_n_bins=5, min_bin_size=0.08),
        OptimalBucketer(variables=cat_features, variables_type="categorical", max_n_bins=5, min_bin_size=0.08),
    )
    bucketer = BucketingProcess(prebucketing_pipeline=prebucketing_pipeline, bucketing_pipeline=bucketing_pipeline)

    expected_probas = np.array([[0.851, 0.149], [0.747, 0.252]])

    run_checks(X, y, bucketer, features, expected_probas)

    # and with features not defined
    features = []
    expected_probas = np.array([[0.88, 0.12], [0.76, 0.24]])
    run_checks(X, y, bucketer, features, expected_probas)


def test_skorecard_with_pipeline_of_buckets(df):
    """Test a workflow, with numerical pipeline of bucketers."""
    X = df.drop("default", axis=1)
    y = df["default"]
    features = ["LIMIT_BAL", "BILL_AMT1"]
    cat_features = [col for col in X.columns if col not in features]

    bucketer = make_pipeline(
        DecisionTreeBucketer(variables=features, max_n_bins=5),
        OrdinalCategoricalBucketer(variables=cat_features, tol=0.05),
    )

    expected_probas = np.array([[0.841, 0.159], [0.738, 0.262]])
    run_checks(X, y, bucketer, features, expected_probas)

    features = []
    expected_probas = np.array([[0.877, 0.122], [0.734, 0.266]])
    run_checks(X, y, bucketer, features, expected_probas)


def test_default_skorecard_class(df):
    """Test a workflow, when no bucketer is defined."""
    X = df.drop("default", axis=1)
    y = df["default"]
    features = ["LIMIT_BAL", "BILL_AMT1"]

    skorecard_model = Skorecard(verbose=0, variables=features)
    skorecard_model.fit(X, y)
    assert isinstance(skorecard_model.bucketing_, BucketingProcess)

    bucketer = None
    expected_probas = np.array([[0.862, 0.138], [0.748, 0.252]])
    run_checks(X, y, bucketer, features, expected_probas)

    bucketer = None
    features = []
    expected_probas = np.array([[0.895, 0.105], [0.752, 0.248]])
    run_checks(X, y, bucketer, features, expected_probas)


def test_random_state(df):
    """Test that all of the random_state shenanigans work as expected."""
    X = df.drop("default", axis=1)
    y = df["default"]
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    # Test assigning a random state
    prebucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=100),
        OrdinalCategoricalBucketer(variables=cat_cols, tol=0.01),
    )
    bucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=100),
        OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=5, min_bin_size=0.08),
    )

    bucketer = BucketingProcess(prebucketing_pipeline=prebucketing_pipeline, bucketing_pipeline=bucketing_pipeline)
    skorecard_model = Skorecard(bucketing=bucketer, variables=num_cols + cat_cols, random_state=411)
    skorecard_model.fit(X, y)
    assert bucketer.pre_pipeline_.named_steps["decisiontreebucketer"].random_state == 411
    assert bucketer.pipeline_.named_steps["decisiontreebucketer"].random_state == 411
    assert skorecard_model.pipeline_.named_steps["model"].random_state == 411

    # Test overwriting a random state
    prebucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=100, random_state=123),
        OrdinalCategoricalBucketer(variables=cat_cols, tol=0.01),
    )
    bucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=100, random_state=123),
        OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=5, min_bin_size=0.08),
    )

    bucketer = BucketingProcess(prebucketing_pipeline=prebucketing_pipeline, bucketing_pipeline=bucketing_pipeline)
    skorecard_model = Skorecard(bucketing=bucketer, variables=num_cols + cat_cols, random_state=411)
    skorecard_model.fit(X, y)
    assert bucketer.pre_pipeline_.named_steps["decisiontreebucketer"].random_state == 411
    assert bucketer.pipeline_.named_steps["decisiontreebucketer"].random_state == 411
    assert skorecard_model.pipeline_.named_steps["model"].random_state == 411

    # Test keeping a random state
    prebucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=100, random_state=123),
        OrdinalCategoricalBucketer(variables=cat_cols, tol=0.01),
    )
    bucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=100, random_state=123),
        OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=5, min_bin_size=0.08),
    )

    bucketer = BucketingProcess(prebucketing_pipeline=prebucketing_pipeline, bucketing_pipeline=bucketing_pipeline)
    skorecard_model = Skorecard(bucketing=bucketer, variables=num_cols + cat_cols, lr_kwargs={"random_state": 411})
    skorecard_model.fit(X, y)
    assert bucketer.pre_pipeline_.named_steps["decisiontreebucketer"].random_state == 123
    assert bucketer.pipeline_.named_steps["decisiontreebucketer"].random_state == 123
    assert skorecard_model.pipeline_.named_steps["model"].random_state == 411

    # Test with Pipeline

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucket_pipe = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=5), OrdinalCategoricalBucketer(variables=cat_cols, tol=0.05)
    )

    skorecard_model = Skorecard(bucketing=bucket_pipe, variables=num_cols + cat_cols, random_state=411)
    skorecard_model.fit(X, y)
    assert skorecard_model.pipeline_.named_steps["bucketer"]["decisiontreebucketer"].random_state == 411
    assert skorecard_model.pipeline_.named_steps["model"].random_state == 411



import numpy as np
import pytest
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder

from skorecard import datasets
from skorecard.bucketers import EqualFrequencyBucketer
from skorecard.linear_model import LogisticRegression


@pytest.fixture()
def X_y():
    """Generate dataframe."""
    X, y = datasets.load_uci_credit_card(return_X_y=True)
    return X, y


@pytest.mark.parametrize("intercept_param", [(True), (False)])
def test_output_dimensions(intercept_param):
    """Test the dimensions of the new attributes."""
    shape_features = (10, 3)

    X = np.random.uniform(size=shape_features)
    y = np.array([1, 0, 0, 1, 0, 0, 1, 0, 0, 1])

    lr = LogisticRegression(fit_intercept=intercept_param).fit(X, y, calculate_stats=True)
    assert lr.p_val_coef_.shape[1] == shape_features[1]
    assert lr.z_coef_.shape[1] == shape_features[1]
    assert len(lr.std_err_coef_) == shape_features[1]
    for statement in [np.isnan(lr.p_val_intercept_), np.isnan(lr.z_intercept_), np.isnan(lr.std_err_intercept_)]:
        assert not intercept_param == statement


def test_results(X_y):
    """Test the actual p-values."""
    expected_approx_p_val_coef_ = np.array([[1.0, 1.0, 0.0, 0.8425]])

    lr = LogisticRegression(fit_intercept=True, penalty=None, solver="sag").fit(*X_y, calculate_stats=True)

    np.testing.assert_array_almost_equal(lr.p_val_coef_, expected_approx_p_val_coef_, decimal=3)


def test_linear_model(X_y):
    """Test OHE sparse matrix compatibility."""
    pipeline = Pipeline(
        [
            ("bucketer", EqualFrequencyBucketer(n_bins=10)),
            ("ohe", OneHotEncoder()),
            ("clf", LogisticRegression(calculate_stats=True)),
        ]
    )
    pipeline.fit(*X_y)
    assert pipeline.named_steps["clf"].p_val_coef_.shape[1] > 0


def test_get_stats(X_y):
    """Test that we get the expected pandas dataframe."""
    lr = LogisticRegression(fit_intercept=True).fit(*X_y)

    with pytest.raises(AssertionError):
        lr.get_stats()

    lr = LogisticRegression(fit_intercept=True).fit(*X_y, calculate_stats=True)

    # We add 1 because of the intercept
    assert lr.get_stats().shape[0] == len(X_y[0].columns) + 1
    assert lr.get_stats().index[0] == "const"

    lr = LogisticRegression(fit_intercept=False, calculate_stats=True).fit(*X_y)
    assert lr.get_stats().fillna(-999)["Coef."][0] == 0
    assert lr.get_stats().fillna(-999)["Std.Err"][0] == -999
    assert lr.get_stats().fillna(-999)["z"][0] == -999
    assert lr.get_stats().fillna(-999)["P>|z|"][0] == -999



import logging
from typing import Dict, List

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from skorecard.features_bucket_mapping import FeaturesBucketMapping
from skorecard.reporting.plotting import PlotBucketMethod
from skorecard.reporting.report import BucketTableMethod, SummaryMethod
from skorecard.utils import BucketingPipelineError, NotBucketObjectError, NotInstalledError
from skorecard.utils.validation import is_fitted

# JupyterDash
try:
    from jupyter_dash import JupyterDash
except ModuleNotFoundError:
    JupyterDash = NotInstalledError("jupyter-dash", "dashboard")

try:
    import dash_bootstrap_components as dbc
except ModuleNotFoundError:
    dbc = NotInstalledError("dash_bootstrap_components", "dashboard")


from skorecard.apps.app_callbacks import add_bucketing_callbacks
from skorecard.apps.app_layout import add_basic_layout


class KeepPandas(BaseEstimator, TransformerMixin):
    """
    Wrapper to keep column names of pandas dataframes in a `scikit-learn` transformer.

    Any scikit-learn transformer wrapped in KeepPandas will return a `pd.DataFrame` on `.transform()`.

    !!! warning
        You should only use `KeepPandas()` when you know for sure `scikit-learn`
        did not change the order of your columns.

    Example:

    ```python
    from skorecard.pipeline import KeepPandas
    from skorecard import datasets
    from skorecard.bucketers import EqualWidthBucketer

    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import StandardScaler

    X, y = datasets.load_uci_credit_card(return_X_y=True)

    bucket_pipeline = make_pipeline(
        KeepPandas(StandardScaler()),
        EqualWidthBucketer(n_bins=5, variables=['LIMIT_BAL', 'BILL_AMT1']),
    )
    bucket_pipeline.fit_transform(X, y)
    ```
    """

    def __init__(self, transformer):
        """Initialize."""
        self.transformer = transformer

        # Warn if there is a chance order of columns are changed
        if isinstance(transformer, Pipeline):
            for step in _get_all_steps(transformer):
                self._check_for_column_transformer(step)
        else:
            self._check_for_column_transformer(transformer)

    def __repr__(self):
        """String representation."""
        return self.transformer.__repr__()

    @staticmethod
    def _check_for_column_transformer(obj):
        msg = "sklearn.compose.ColumnTransformer can change the order of columns"
        msg += ", be very careful when using with KeepPandas()"
        if type(obj).__name__ == "ColumnTransformer":
            logging.warning(msg)

    def fit(self, X, y=None, *args, **kwargs):
        """Fit estimator."""
        assert isinstance(X, pd.DataFrame)
        self.columns_ = list(X.columns)
        self.transformer.fit(X, y, *args, **kwargs)
        return self

    def transform(self, X, *args, **kwargs):
        """Transform X."""
        check_is_fitted(self)
        new_X = self.transformer.transform(X, *args, **kwargs)
        return pd.DataFrame(new_X, columns=self.columns_)

    def get_feature_names(self):
        """Return estimator feature names."""
        check_is_fitted(self)
        return self.columns_


def find_bucketing_step(pipeline: Pipeline, identifier: str = "bucketingprocess"):
    """
    Finds a specific step in a sklearn Pipeline that has a 'name' attribute equalling 'identifier'.

    This is useful to extract certain steps from a pipeline, f.e. a BucketingProcess.

    Args:
        pipeline (sklearn.pipeline.Pipeline): sklearn pipeline
        identifier (str): the attribute used to find the pipeline step

    Returns:
        index (int): position of bucketing step in pipeline.steps
    """
    # Find the bucketing pipeline step
    bucket_pipes = [s for s in pipeline.steps if getattr(s[1], "name", "") == identifier]

    # Raise error if missing
    if len(bucket_pipes) == 0:
        msg = """
        Did not find a bucketing pipeline step. Identity the bucketing pipeline step
        using skorecard.pipeline.make_bucketing_pipeline or skorecard.pipeline.make_prebucketing_pipeline.

        Note that the pipeline should always have a skorecard.pipeline.make_prebucketing_pipeline defined.
        If you do not need prebucketing simply leave it empty.

        Example:

        ```python
        from sklearn.pipeline import make_pipeline
        from skorecard.pipeline import make_bucketing_pipeline, make_prebucketing_pipeline

        pipeline = make_pipeline(
            make_prebucketing_pipeline(),
            make_bucketing_pipeline(
                    OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
                    OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
            )
        )
        ```
        """
        raise AssertionError(msg)

    if len(bucket_pipes) > 1:
        msg = """
        You need to identity only the bucketing step,
        using skorecard.pipeline.make_bucketing_pipeline and skorecard.pipeline.make_prebucketing_pipeline only once.

        Example:

        ```python
        from skorecard.pipeline import make_bucketing_pipeline
        bucket_pipeline = make_bucketing_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        )
        ```
        """
        raise AssertionError(msg)

    index_bucket_pipeline = pipeline.steps.index(bucket_pipes[0])
    return index_bucket_pipeline


def get_features_bucket_mapping(pipe: Pipeline) -> FeaturesBucketMapping:
    """Get feature bucket mapping from a sklearn pipeline object.

    ```python
    from skorecard import datasets
    from skorecard.bucketers import EqualWidthBucketer, OrdinalCategoricalBucketer
    from skorecard.pipeline import get_features_bucket_mapping

    from sklearn.pipeline import Pipeline, make_pipeline
    from sklearn.preprocessing import OneHotEncoder
    from sklearn.metrics import roc_auc_score
    from sklearn.linear_model import LogisticRegression

    X, y = datasets.load_uci_credit_card(return_X_y=True)

    bucket_pipeline = make_pipeline(
        EqualWidthBucketer(n_bins=5, variables=['LIMIT_BAL', 'BILL_AMT1']),
        OrdinalCategoricalBucketer(variables=['EDUCATION', 'MARRIAGE'])
    )

    pipe = Pipeline([
        ('bucketing', bucket_pipeline),
        ('one-hot-encoding', OneHotEncoder()),
        ('lr', LogisticRegression())
    ])

    pipe.fit(X, y)
    features_bucket_mapping = get_features_bucket_mapping(pipe)
    ```

    Args:
        pipe (Pipeline): fitted scikitlearn pipeline with bucketing transformers

    Returns:
        FeaturesBucketMapping: skorecard class with the bucket info
    """
    assert isinstance(pipe, BaseEstimator)

    features_bucket_mapping = {}
    for step in _get_all_steps(pipe):
        check_is_fitted(step)
        if hasattr(step, "features_bucket_mapping_"):
            features_bucket_mapping.update(step.features_bucket_mapping_.as_dict())

    assert (
        len(features_bucket_mapping) > 0
    ), "pipeline does not have any fitted skorecard bucketer. Update the pipeline or fit(X,y) first"
    return FeaturesBucketMapping(features_bucket_mapping)


def _get_all_steps(pipeline: Pipeline) -> List:
    """
    Returns a list of steps in a sklearn pipeline.

    Args:
        pipeline (Pipeline): A scikitlearn pipeline.
    """
    steps = []
    for step in pipeline.steps:
        if isinstance(step, tuple):
            step = step[1]
        if hasattr(step, "steps"):
            steps += _get_all_steps(step)
        else:
            steps.append(step)
    return steps


class SkorecardPipeline(Pipeline, PlotBucketMethod, BucketTableMethod, SummaryMethod):
    """
    A sklearn Pipeline with several attribute and methods added.

    This Pipeline of bucketers behaves more like a bucketer and adds:

    - `.summary()`: See which columns are bucketed
    - `.plot_bucket()`: Plot buckets of a column
    - `.bucket_table()`: Table with buckets of a column
    - `.save_to_yaml()`: Save information necessary for bucketing to a YAML file
    - `.features_bucket_mapping_`: Access bucketing information
    - `.fit_interactive()`: Edit fitted buckets interactively in a dash app

    ```python
    from skorecard.pipeline.pipeline import SkorecardPipeline
    from skorecard.bucketers import DecisionTreeBucketer, OrdinalCategoricalBucketer
    from skorecard import datasets

    pipe = SkorecardPipeline([
        ('decisiontreebucketer', DecisionTreeBucketer(variables = ["LIMIT_BAL", "BILL_AMT1"],max_n_bins=5)),
        ('ordinalcategoricalbucketer', OrdinalCategoricalBucketer(variables = ["EDUCATION", "MARRIAGE"], tol =0.05)),
    ])

    df = datasets.load_uci_credit_card(as_frame=True)
    features = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]
    X = df[features]
    y = df["default"].values

    pipe.fit(X, y)
    pipe.bucket_table('LIMIT_BAL')
    ```
    """

    def __init__(self, steps, *, memory=None, verbose=False):
        """
        Wraps sklearn Pipeline.
        """
        super().__init__(steps=steps, memory=memory, verbose=verbose)
        self._check_pipeline_all_bucketers(self)
        self._check_pipeline_duplicated_columns(self)

    @property
    def features_bucket_mapping_(self):
        """
        Retrieve features bucket mapping.
        """
        check_is_fitted(self.steps[-1][1])
        return get_features_bucket_mapping(Pipeline(self.steps))

    @property
    def bucket_tables_(self):
        """
        Retrieve bucket tables.

        Used by .bucket_table()
        """
        check_is_fitted(self.steps[-1][1])
        bucket_tables = dict()
        for step in self.steps:
            bucket_tables.update(step[1].bucket_tables_)
        return bucket_tables

    @property
    def summary_dict_(self) -> Dict:
        """
        Retrieve summary_dicts and combine.

        Used by .summary()
        """
        summary_dict = {}
        for step in self.steps:
            summary_dict.update(step[1].summary_dict_)
        return summary_dict

    def save_yml(self, fout):
        """
        Save the features bucket to a yaml file.

        Args:
            fout: file output
        """
        check_is_fitted(self.steps[-1][1])
        self.features_bucket_mapping_.save_yml(fout)

    @staticmethod
    def _check_pipeline_duplicated_columns(pipeline: Pipeline) -> None:
        """
        Check that the pipeline has no duplicated columns.

        This check only works on fitted pipelines!
        """
        assert isinstance(pipeline, Pipeline)

        bucketers_vars = []
        bucketers_on_all = []
        bucketers_with_vars = []

        for step in _get_all_steps(pipeline):
            if is_fitted(step):
                if hasattr(step, "variables_"):
                    if len(step.variables_) == 0:
                        bucketers_vars += ["**all**"]
                        bucketers_on_all += [step]
                    else:
                        bucketers_vars += step.variables_
                        bucketers_with_vars += [step]
            else:
                if hasattr(step, "variables"):
                    if len(step.variables) == 0:
                        bucketers_vars += ["**all**"]
                        bucketers_on_all += [step]
                    else:
                        bucketers_vars += step.variables
                        bucketers_with_vars += [step]

        if len(list(set(bucketers_vars))) > 1 and "**all**" in list(set(bucketers_vars)):
            msg = "A SkorecardPipeline should bucket each feature only once.\n"
            msg += f"These bucketers bucket all features: {bucketers_on_all}\n"
            msg += f"While these bucket specific ones: {bucketers_with_vars}\n"
            msg += "This means some features would have been bucketed sequentially."
            msg += "To solve this, either use a BucketingProcess, or remove the duplicates from one of the bucketers."
            msg += "Remember that if you don't specify 'variables', a bucketer will bucket all columns."
            raise BucketingPipelineError(msg)

        if len(set(bucketers_vars)) != len(bucketers_vars):
            values, counts = np.unique(bucketers_vars, return_counts=True)
            duplicates = list(set(values[counts > 1]))

            msg = "A SkorecardPipeline should bucket each feature only once. "
            msg += f"The features {duplicates} appear in multiple bucketers, "
            msg += "meaning they would have been bucketed sequentially."
            msg += "To solve this, either use a BucketingProcess, or remove the duplicates from one of the bucketers."
            msg += "Remember that if you don't specify 'variables', a bucketer will bucket all columns."
            raise BucketingPipelineError(msg)

    @staticmethod
    def _check_pipeline_all_bucketers(pipeline: Pipeline) -> None:
        """
        Ensure all specified bucketing steps are skorecard bucketers.

        Args:
            pipeline: scikit-learn pipeline.
        """
        assert isinstance(pipeline, Pipeline)

        for step in _get_all_steps(pipeline):
            if all(x not in str(type(step)) for x in ["bucketing_process", "skorecard.bucketers"]):
                msg = "All bucketing steps must be skorecard bucketers."
                msg += f"Remove {step} from the pipeline."
                raise NotBucketObjectError(msg)

    def fit_interactive(self, X, y=None, mode="external"):
        """
        Fit a bucketer and then interactively edit the fit using a dash app.

        Note we are using a [jupyterdash](https://medium.com/plotly/introducing-jupyterdash-811f1f57c02e) app,
        which supports 3 different modes:

        - 'external' (default): Start dash server and print URL
        - 'inline': Start dash app inside an Iframe in the jupyter notebook
        - 'jupyterlab': Start dash app as a new tab inside jupyterlab

        """
        # We need to make sure we only fit if not already fitted
        # This prevents a user losing manually defined boundaries
        # when re-running .fit_interactive()
        if not is_fitted(self):
            self.fit(X, y)

        self.app = JupyterDash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        add_basic_layout(self)
        add_bucketing_callbacks(self, X, y)
        self.app.run_server(mode=mode)

    def _update_column_fit(self, X, y, feature, special, splits, right, generate_summary=False):
        """
        Extract out part of the fit for a column.

        Useful when we want to interactively update the fit.
        """
        for step in self.steps:
            if feature in step[1].variables:
                step[1]._update_column_fit(
                    X=X,
                    y=y,
                    feature=feature,
                    special=special,
                    splits=splits,
                    right=right,
                    generate_summary=generate_summary,
                )


def to_skorecard_pipeline(pipeline: Pipeline) -> SkorecardPipeline:
    """
    Transform a scikit-learn Pipeline to a SkorecardPipeline.

    A SkorecardPipeline is a normal scikit-learn pipeline with some extra methods and attributes.

    Example:

    ```python
    from skorecard.pipeline.pipeline import SkorecardPipeline, to_skorecard_pipeline
    from skorecard.bucketers import DecisionTreeBucketer, OrdinalCategoricalBucketer
    from skorecard import datasets

    from sklearn.pipeline import make_pipeline

    pipe = make_pipeline(
        DecisionTreeBucketer(variables = ["LIMIT_BAL", "BILL_AMT1"],max_n_bins=5),
        OrdinalCategoricalBucketer(variables = ["EDUCATION", "MARRIAGE"], tol =0.05)
    )
    sk_pipe = to_skorecard_pipeline(pipe)

    df = datasets.load_uci_credit_card(as_frame=True)

    features = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]
    X = df[features]
    y = df["default"].values
    ```

    Args:
        pipeline (Pipeline): `scikit-learn` pipeline instance.

    Returns:
        pipeline (skorecard.pipeline.SkorecardPipeline): modified pipeline instance.
    """
    assert isinstance(pipeline, Pipeline)
    if isinstance(pipeline, SkorecardPipeline):
        return pipeline
    else:
        return SkorecardPipeline(steps=pipeline.steps, memory=pipeline.memory, verbose=pipeline.verbose)



import numpy as np
import pandas as pd
import scipy

from skorecard.utils.exceptions import DimensionalityError


def reshape_1d_to_2d(x):
    """Converts/reshapes the input x to a numpy array o (n,1).

    Args:
        x: list, numpy array, pandas dataframe (one column only), pandas series

    Returns: numpy array of shape (n,1)
    """
    X_array = None
    if isinstance(x, list):
        X_array = np.array(x)
    if isinstance(x, np.ndarray):
        if x.ndim > 1 and x.shape[1] > 1:
            raise DimensionalityError("Expected one column only. Can't reshape")
        X_array = x
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1:
            X_array = x.values.flatten()
        else:
            raise DimensionalityError("Expected one column only. Can't reshape")
    if isinstance(x, pd.core.series.Series):
        X_array = x.values

    return X_array.reshape(-1, 1)


def convert_sparse_matrix(x):
    """
    Converts a sparse matrix to a numpy array.

    This can prevent problems arising from, e.g. OneHotEncoder.

    Args:
        x: numpy array, sparse matrix

    Returns:
        numpy array of x
    """
    if scipy.sparse.issparse(x):
        return x.toarray()
    else:
        return x



import numpy as np
import pandas as pd
import pytest
from sklearn.pipeline import make_pipeline

from skorecard.bucket_mapping import BucketMapping, MissingDict, merge_bucket_mapping
from skorecard.bucketers.bucketers import DecisionTreeBucketer, EqualWidthBucketer
from skorecard.features_bucket_mapping import FeaturesBucketMapping, merge_features_bucket_mapping


def test_bucket_mapping_numerical():
    """Tests numerical transforms."""
    bm = BucketMapping("testfeat", "numerical", map=[1, 3, 5], right=True)
    assert bm.labels == {0: "(-inf, 1.0]", 1: "(1.0, 3.0]", 2: "(3.0, 5.0]", 3: "(5.0, inf]", -1: "Missing"}
    assert all(bm.transform([1, 3, 5, 7]) == np.array([0, 1, 2, 3]))

    # with right=False
    bm = BucketMapping("testfeat", "numerical", map=[1, 3, 5], right=False)
    assert bm.labels == {0: "[-inf, 1.0)", 1: "[1.0, 3.0)", 2: "[3.0, 5.0)", 3: "[5.0, inf)", -1: "Missing"}
    assert all(bm.transform([1, 3, 5, 7]) == np.array([1, 2, 3, 3]))

    x = [0, 1, 2, 3, 4, 5]
    bucket = BucketMapping("feature1", "numerical", map=[3, 4])
    assert all(np.equal(bucket.transform(x), np.digitize(x, [3, 4], right=True)))
    # array([0, 0, 0, 0, 1, 2])

    x = [0, 1, 2, 3, 4, 5]
    bucket = BucketMapping("feature1", "numerical", map=[2, 3, 4])
    assert all(np.equal(bucket.transform(x), np.digitize(x, [2, 3, 4], right=True)))
    # array([0, 0, 0, 1, 2, 3])

    x = [0, 1, np.nan, 3, 4, 5]
    bucket = BucketMapping("feature1", "numerical", map=[2, 3, 4])
    assert np.allclose(bucket.transform(x), np.array([0, 0, -1, 1, 2, 3]), equal_nan=True)
    assert "Missing" in bucket.labels.values()


def test_bucket_mapping_categorical():
    """
    Tests categorical transforms.
    """
    # Make sure that the map outputs start at 0 and are incremental. Because it is skipping 2,it will raise an exception
    with pytest.raises(ValueError):
        bucket = BucketMapping("feature1", "categorical", map={"car": 0, "truck": 0, "boat": 2})

    x = ["car", "motorcycle", "boat", "truck", "truck"]
    bucket = BucketMapping("feature1", "categorical", map={})
    other_category_encoding = -2
    assert all(np.equal(bucket.transform(x), [other_category_encoding] * 5))

    # Empty map with NA's
    x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
    bucket = BucketMapping("feature1", "categorical", map={})
    other_category_encoding = -2
    missing_cat = -1
    assert all(np.equal(bucket.transform(x), [other_category_encoding] * 5 + [missing_cat]))

    # # Limited map
    x = ["car", "motorcycle", "boat", "truck", "truck"]
    bucket = BucketMapping("feature1", "categorical", map={"car": 0, "truck": 0})
    other_category_encoding = -2
    reference = [0, other_category_encoding, other_category_encoding, 0, 0]
    assert all(np.equal(bucket.transform(x), reference))

    # Limited map with NA's
    x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
    bucket = BucketMapping("feature1", "categorical", map={"car": 0, "truck": 0})
    other_category_encoding = -2
    missing_cat = -1
    reference = [0, other_category_encoding, other_category_encoding, 0, 0, missing_cat]
    assert all(np.equal(bucket.transform(x), reference))

    # Specials defined
    a = BucketMapping("feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]})
    assert a.transform([np.nan]) == -1

    a = BucketMapping(
        feature_name="EDUCATION",
        type="categorical",
        missing_bucket=None,
        other_bucket=None,
        map={2: 0, 3: 1, 5: 2, 6: 3, 4: 4, 0: 5},
        right=True,
        specials={"ed 0": [1]},
    )
    assert a.transform([1]) == -3

    # Make sure dtypes is OK
    a = BucketMapping(
        feature_name="EDUCATION",
        type="categorical",
        missing_bucket=None,
        other_bucket=None,
        map={2: 0, 1: 1, 3: 2},
        right=True,
        specials={},
    )
    # number of bits in .astype(int) differs per OS
    assert a.transform([1, 2, 3]).dtype in [np.dtype("int32"), np.dtype("int64")]


def test_cat_other_bucket():
    """
    Test using 'other_bucket' parameter.
    """
    with pytest.raises(AssertionError):
        bucket = BucketMapping("feature1", "categorical", map={"car": 0, "truck": 0}, other_bucket=1)

    bucket = BucketMapping("feature1", "categorical", map={"car": 0, "truck": 1}, other_bucket=1)
    assert bucket.labels == {0: "car", 1: "truck | Other", -1: "Missing"}
    assert bucket.transform(["bla"]).tolist() == [1]

    bucket = BucketMapping("feature1", "categorical", map={"car": 0, "truck": 1})
    assert bucket.transform(["bla"]).tolist() == [-2]


def test_reserved_names():
    """
    Test using reserved names.
    """
    a = BucketMapping("feature1", "categorical", map={"other": 0, "truck": 1})
    assert a.transform(["other"]).tolist() == [0]
    assert a.transform(["sdfsfsa"]).tolist() == [-2]


def test_get_labels():
    """Make sure nicely formatting is returned."""
    bucket = BucketMapping("feature1", "numerical", map=[1, 3, 4], right=True)
    assert bucket.labels == {0: "(-inf, 1.0]", 1: "(1.0, 3.0]", 2: "(3.0, 4.0]", 3: "(4.0, inf]", -1: "Missing"}

    bucket = BucketMapping("feature1", "numerical", map=[1, 3, 4], right=False)
    assert bucket.labels == {0: "[-inf, 1.0)", 1: "[1.0, 3.0)", 2: "[3.0, 4.0)", 3: "[4.0, inf)", -1: "Missing"}

    bucket = BucketMapping("feature1", "numerical", map=[1], right=True)
    assert bucket.labels == {0: "(-inf, 1.0]", 1: "(1.0, inf]", -1: "Missing"}

    bucket = BucketMapping("feature1", "numerical", map=[1], right=False)
    assert bucket.labels == {0: "[-inf, 1.0)", 1: "[1.0, inf)", -1: "Missing"}

    bucket = BucketMapping("feature1", "numerical", map=[], right=True)
    assert bucket.labels == {0: "(-inf, inf]", -1: "Missing"}

    bucket = BucketMapping("feature1", "numerical", map=[], right=False)
    assert bucket.labels == {0: "[-inf, inf)", -1: "Missing"}


def test_specials_numerical():
    """Test that the specials are put in a special bin."""
    # test that a special case
    x = [0, 1, 2, 3, 4, 5, 2]
    bucket = BucketMapping("feature1", "numerical", map=[3, 4], specials={"=2": [2]})
    assert all(np.equal(bucket.transform(x), np.array([0, 0, -3, 0, 1, 2, -3])))

    # test that calling transform again does not change the labelling
    assert bucket.labels == {0: "(-inf, 3.0]", 1: "(3.0, 4.0]", 2: "(4.0, inf]", -1: "Missing", -3: "Special: =2"}
    assert all(bucket.transform(x) == bucket.transform(x))
    assert bucket.labels == {0: "(-inf, 3.0]", 1: "(3.0, 4.0]", 2: "(4.0, inf]", -1: "Missing", -3: "Special: =2"}

    # Test that if special is not in x, nothing happens
    x = [0, 1, 2, 3, 4, 5]
    bucket = BucketMapping("feature1", "numerical", map=[3, 4], specials={"=6": [6]})
    assert all(np.equal(bucket.transform(x), np.digitize(x, [3, 4], right=True)))

    # Test multiple specials
    bucket = BucketMapping("feature1", "numerical", map=[3, 4], specials={"=2": [2], "=3or4": [3, 4]})
    assert bucket.labels == {
        0: "(-inf, 3.0]",
        1: "(3.0, 4.0]",
        2: "(4.0, inf]",
        -1: "Missing",
        -3: "Special: =2",
        -4: "Special: =3or4",
    }


def test_labels():
    """Test that the labels are correct in different scenarios."""
    x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
    bucket = BucketMapping("feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]})
    bins = pd.Series(bucket.transform(x))

    in_series = pd.Series(x)

    labels = bins.map(bucket.labels)

    # labels with specials for categorical are not so nice
    # we cannot include the special name because item become part of self.map
    assert labels[in_series == "truck"].equals(labels[labels == "Special: is truck"])
    assert labels[(in_series.isin(["car", "boat"]))].equals(labels[labels == "boat, car"])

    # test with numerical categories
    # Limited map with NA's
    x = [310, 311, 312, 313, 313, np.nan]
    bucket = BucketMapping("feature1", "categorical", map={310: 0, 311: 1, 312: 2}, specials={"is 313": [313]})
    bins = pd.Series(bucket.transform(x))

    in_series = pd.Series(x)

    labels = bins.map(bucket.labels)

    assert labels[in_series == 313].equals(labels[labels == "Special: is 313"])
    assert labels[in_series == 311].equals(labels[labels == "311"])
    assert labels[in_series.isna()].equals(labels[labels == "Missing"])

    # test numerical labels
    # note that .transform() is not necessary to build the labels.
    x = [0, 1, 2, 3, 4, 5, 2, np.nan]
    bucket = BucketMapping("feature1", "numerical", map=[3, 4], specials={"=2": [2]})
    assert bucket.labels == {0: "(-inf, 3.0]", 1: "(3.0, 4.0]", 2: "(4.0, inf]", -1: "Missing", -3: "Special: =2"}
    bins = pd.Series(bucket.transform(x))
    in_series = pd.Series(x)

    labels = bins.map(bucket.labels)

    assert labels[(in_series <= 3) & (in_series != 2)].equals(labels[labels == "(-inf, 3.0]"])
    assert labels[(in_series <= 4) & (in_series > 3)].equals(labels[labels == "(3.0, 4.0]"])
    assert labels[in_series == 2].equals(labels[labels == "Special: =2"])
    assert labels[in_series > 4].equals(labels[labels == "(4.0, inf]"])


def test_error_is_raised_if_wrong_specials():
    """Self explanatory."""
    # Test that is the values of the dictionary
    with pytest.raises(AssertionError):
        BucketMapping("feature1", "numerical", map=[3, 4], specials={"special": 2})
    #
    with pytest.raises(AssertionError):
        BucketMapping("feature1", "numerical", map=[3, 4], specials={0: [2]})


def test_features_bucket_mapping():
    """
    Test construction feature bucketing mapping.
    """
    a = BucketMapping("testfeat1", "numerical", map=[1, 3, 5], right=True)
    b = BucketMapping("testfeat2", "numerical", map=[1, 3, 5], right=False)
    fbm = FeaturesBucketMapping([a, b])
    assert fbm.get("testfeat1") == a
    assert fbm.get("testfeat2") == b
    assert len(fbm) == 2

    # TODO: Remove or include
    # features_dict = {
    #     'feature1': {'feature_name': 'feature1',
    #         'type': 'numerical',
    #         'map': [2, 3, 4, 5],
    #         'right': True},
    #     'feature2': {'feature_name': 'feature2',
    #         'type': 'numerical',
    #         'map': [5, 6, 7, 8],
    #         'right': True}
    # }
    # assert FeaturesBucketMapping(features_dict) == FeaturesBucketMapping().load_dict(features_dict)


def test_merge_bucket_mapping_numerical():
    """
    Test merging two numerical bucketmappings.
    """
    a = BucketMapping(
        feature_name="LIMIT_BAL",
        type="numerical",
        map=[
            25000.0,
            45000.0,
            55000.0,
            75000.0,
            85000.0,
            105000.0,
            145000.0,
            175000.0,
            225000.0,
            275000.0,
            325000.0,
            385000.0,
        ],
        right=False,
    )

    b = BucketMapping(
        feature_name="LIMIT_BAL",
        type="numerical",
        map=[1, 2, 4, 6, 7, 9, 10, 12],
        right=False,
    )

    c = merge_bucket_mapping(a, b)

    x = [25000, 45000]
    assert all(c.transform(x) == b.transform(a.transform(x)))

    x = [
        -1,
        0,
        24999,
        25000.0,
        45000.0,
        55000.0,
        75000.0,
        85000.0,
        105000.0,
        145000.0,
        175000.0,
        225000.0,
        275000.0,
        325000.0,
        385000.0,
        400000,
    ]
    assert all(c.transform(x) == b.transform(a.transform(x)))

    # Now first bucketer with right=True, second one with right=False
    a = BucketMapping("testfeat", "numerical", map=[1, 3, 5], right=True)
    assert all(a.transform([1, 3, 5, 7]) == np.array([0, 1, 2, 3]))
    b = BucketMapping("testfeat", "numerical", map=[1, 3], right=False)
    assert all(b.transform([1, 3, 5, 7]) == np.array([1, 2, 2, 2]))
    c = merge_bucket_mapping(a, b)
    x = [-1, 0, 0.99, 1, 1.001, 3, 3.1, 4.9, 5, 5.1]
    assert all(c.transform(x) == b.transform(a.transform(x)))

    # Now with specials
    a = BucketMapping(
        feature_name="LIMIT_BAL",
        type="numerical",
        map=[
            25000.0,
            45000.0,
            55000.0,
            75000.0,
            85000.0,
            105000.0,
            145000.0,
            175000.0,
            225000.0,
            275000.0,
            325000.0,
            385000.0,
        ],
        right=False,
        specials={"=45000.0": [45000.0]},
    )

    b = BucketMapping(
        feature_name="LIMIT_BAL",
        type="numerical",
        map=[1, 2, 4, 6, 7, 9, 10, 12],
        right=False,
        specials={"=45000.0": [14]},
    )

    c = merge_bucket_mapping(a, b)
    assert all(c.transform(x) == b.transform(a.transform(x)))


def test_merge_bucket_mapping_categorical():
    """
    Test merging two categorical bucketmappings.
    """
    a = BucketMapping(
        feature_name="EDUCATION",
        type="categorical",
        map={
            0: 6,
            1: 0,
            2: 1,
            3: 2,
            4: 5,
            5: 3,
            6: 4,
        },
        right=True,
        specials={},
    )
    b = BucketMapping(
        feature_name="EDUCATION",
        type="categorical",
        map={
            0: 0,
            1: 2,
            2: 1,
            3: 0,
            4: 0,
            5: 0,
            6: 0,
        },
        right=False,
        specials={},
    )

    correct_map = {0: 0, 1: 0, 2: 2, 3: 1, 4: 0, 5: 0, 6: 0}
    c = merge_bucket_mapping(a, b)
    assert c.map == correct_map

    x = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    assert all(c.transform(x) == b.transform(a.transform(x)))

    # another case
    a = BucketMapping(
        feature_name="EDUCATION",
        type="categorical",
        map={2: 0, 1: 1, 3: 2},
        right=True,
        specials={},
    )
    b = BucketMapping(
        feature_name="EDUCATION",
        type="categorical",
        map={3: 0, 1: 0, 2: 1, 0: 2},
        right=False,
        specials={},
    )

    c = merge_bucket_mapping(a, b)
    x = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    assert all(c.transform(x) == b.transform(a.transform(x)))


def test_missing_dict():
    """
    Test MissingDict.
    """
    s = pd.Series(["cat", "dog", np.nan, "rabbit"])
    a = {"cat": "kitten", "dog": "puppy"}
    a = MissingDict(a)
    a.set_missing_value("bye")
    assert all(s.map(a) == pd.Series(["kitten", "puppy", "bye", "bye"]))
    assert pd.isnull(s.map(a, na_action="ignore")[2])


def test_missing_bucket():
    """
    Test missing values are treated properly.
    """
    # missing_bucket='separate'
    a = BucketMapping("feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]})
    x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
    assert a.transform(x)[5] == -1

    # missing_bucket=<bucket that exists>
    for missing_bucket in [0]:
        a = BucketMapping(
            "feature1",
            "categorical",
            map={"car": 0, "boat": 0},
            specials={"is truck": ["truck"]},
            missing_bucket=missing_bucket,
        )
        x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
        assert a.transform(x)[5] == missing_bucket

    # missing_bucket=<bucket that does not exist in map>
    # should raise an error!
    with pytest.raises(AssertionError):
        a = BucketMapping(
            "feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]}, missing_bucket=7
        )

    # numerical. missing_bucket='separate'
    a = BucketMapping("testfeat", "numerical", map=[1, 3, 5], right=True)
    x = [1, 3, 5, np.nan, 7]
    assert all(a.transform(x) == np.array([0, 1, 2, -1, 3]))

    # numerical. missing_bucket=1
    for missing_bucket in range(3):
        a = BucketMapping("testfeat", "numerical", missing_bucket=missing_bucket, map=[1, 3, 5], right=True)
        x = [1, 3, 5, np.nan, 7]
        assert all(a.transform(x) == np.array([0, 1, 2, missing_bucket, 3]))

    # numerical missing_bucket=<bucket that does not exist in map>
    # should raise an error!
    with pytest.raises(AssertionError):
        a = BucketMapping("testfeat", "numerical", missing_bucket=4, map=[1, 3, 5], right=True)


def test_other_bucket():
    """
    Tests other_bucket param.
    """
    # other_bucket=<bucket that exists>
    for other_bucket in [0]:
        a = BucketMapping(
            "feature1",
            "categorical",
            map={"car": 0, "boat": 0},
            specials={"is truck": ["truck"]},
            other_bucket=other_bucket,
        )
        x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
        assert a.transform(x)[2] == other_bucket

    # other_bucket=<bucket that does not exist in map>
    # should raise an error!
    with pytest.raises(AssertionError):
        a = BucketMapping(
            "feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]}, other_bucket=7
        )


def test_missing_other_bucket():
    """
    Test transforms when both other_bucket and missing_bucket are defined.
    """
    a = BucketMapping("feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]})
    x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
    assert a.labels == {0: "boat, car", -2: "Other", -1: "Missing", -3: "Special: is truck"}
    assert a.transform(x)[1] == -2  # other
    assert a.transform(x)[4] == -3  # special
    assert a.transform(x)[5] == -1  # missing

    # other_bucket=<bucket that exists>
    for bucket in [0]:
        a = BucketMapping(
            "feature1",
            "categorical",
            map={"car": 0, "boat": 0},
            specials={"is truck": ["truck"]},
            other_bucket=bucket,
            missing_bucket=bucket,
        )
        x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
        assert a.transform(x)[2] == bucket
        assert a.transform(x)[5] == bucket

    # other_bucket=<bucket that does not exist in map>
    # should raise an error!
    with pytest.raises(AssertionError):
        a = BucketMapping(
            "feature1",
            "categorical",
            map={"car": 0, "boat": 0},
            specials={"is truck": ["truck"]},
            other_bucket=7,
            missing_bucket=7,
        )


def test_merge_cats_with_missing():
    """
    Testing merges.
    """
    # Test wrong 'other_bucket' set
    a = BucketMapping("feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]})
    b = BucketMapping(
        "feature1", "categorical", map={0: 0, 1: 1, 3: 1, 2: 0}, other_bucket=1, specials={"is truck": [-3]}
    )
    with pytest.raises(AssertionError):
        c = merge_bucket_mapping(a, b)

    # Test wrong 'missing_bucket' set
    a = BucketMapping("feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]})
    b = BucketMapping(
        "feature1", "categorical", map={0: 0, 1: 1, 3: 1, 2: 0}, missing_bucket=1, specials={"is truck": [-3]}
    )
    with pytest.raises(AssertionError):
        c = merge_bucket_mapping(a, b)

    a = BucketMapping("feature1", "categorical", map={"car": 0, "boat": 0}, specials={"is truck": ["truck"]})
    b = BucketMapping(
        "feature1", "categorical", map={0: 0, -1: 0, -2: 0}, missing_bucket=0, specials={"is truck": [-3]}
    )
    c = merge_bucket_mapping(a, b)
    x = ["car", "motorcycle", "boat", "truck", "truck", np.nan]
    assert all(c.transform(x) == b.transform(a.transform(x)))

    x = [0, 1, 2, 3, 4, 5, 6, 7, 8]
    assert all(c.transform(x) == b.transform(a.transform(x)))

    # first bucket with lots of cats
    # second with very few
    a = BucketMapping(
        "feature1", "categorical", map={"car": 0, "boat": 1, "motorcycle": 2}, specials={"is truck": ["truck"]}
    )
    b = BucketMapping(
        "feature1",
        "categorical",
        map={0: 0, 1: 0, 2: 1},
        other_bucket=0,
        missing_bucket=0,
        specials={"is truck": [-3]},
    )

    c = merge_bucket_mapping(a, b)
    x = ["car", "motorcycle", "boat", "truck", "truck", np.nan, "something else"]
    assert all(c.transform(x) == b.transform(a.transform(x)))


def test_merge_features_bucket_mapping():
    """
    Tests merging two sets.
    """
    a = BucketMapping("testfeat", "numerical", map=[1, 3, 5], right=True)
    b = BucketMapping("testfeat", "numerical", map=[1, 3], right=False)
    c = merge_bucket_mapping(a, b)

    af = FeaturesBucketMapping([a])
    bf = FeaturesBucketMapping([b])

    assert FeaturesBucketMapping([c]) == merge_features_bucket_mapping(af, bf)

    # now with new keys in either
    a = BucketMapping("testfeat", "numerical", map=[1, 3, 5], right=True)
    b = BucketMapping("testfeat", "numerical", map=[1, 3], right=False)
    c = BucketMapping("new_feat", "numerical", map=[5, 6], right=False)

    af = FeaturesBucketMapping([a, c])
    bf = FeaturesBucketMapping([b])

    assert FeaturesBucketMapping([merge_bucket_mapping(a, b), c]) == merge_features_bucket_mapping(af, bf)


def test_merge_buckets_on_data(df):
    """
    Test on actual data.
    """
    df = df.drop(columns=["pet_ownership", "BILL_AMT1"])
    y = df["default"]
    X = df.drop(columns=["default"])

    buck_a = EqualWidthBucketer(n_bins=100)
    buck_b = DecisionTreeBucketer()

    pipe = make_pipeline(buck_a, buck_b)
    pipe.fit(X, y)

    a = pipe.steps[0][1].features_bucket_mapping_
    b = pipe.steps[1][1].features_bucket_mapping_
    c = merge_features_bucket_mapping(a, b)

    for feature in X.columns:
        assert all(c.get(feature).transform(X[feature]) == pipe.transform(X)[feature].values)


def test_item_assignment():
    """
    Test if assignment bucket mappings by feature name works properly.
    """
    bm = FeaturesBucketMapping()
    a = BucketMapping("feature1", "categorical", map={310: 0, 311: 1, 312: 2}, specials={"is 313": [313]})
    b = BucketMapping("feature1", "categorical", map={310: 0, 311: 1, 312: 2}, specials={"is 999": [313]})

    bm.append(a)
    bm["feature1"] = b
    assert len(bm) == 1
    assert bm.get("feature1") == b



import numpy as np
import pytest

from skorecard import datasets
from skorecard.bucketers import DecisionTreeBucketer, EqualFrequencyBucketer, EqualWidthBucketer, OptimalBucketer


@pytest.fixture()
def df():
    """Generate dataframe."""
    return datasets.load_uci_credit_card(as_frame=True)


def test_specials_tree_bucketer(df):
    """Test that when adding specials,the binner performs as expected.

    Context: special values should be binned in their own bin.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    # Because 2 special buckets are defined, the decision tree
    # will be fitted with max_leaf_nodes=1. This will create a crash in the sklearn implementation.
    # In this case, Skorecard raises an exception with a clear recommendation to the user when the fit method is called.
    tbt = DecisionTreeBucketer(variables=["LIMIT_BAL", "BILL_AMT1"], random_state=1, max_n_bins=3, specials=specials)
    with pytest.raises(ValueError):
        tbt.fit_transform(X, y)

    tbt = DecisionTreeBucketer(variables=["LIMIT_BAL", "BILL_AMT1"], random_state=1, max_n_bins=5, specials=specials)
    X_bins = tbt.fit_transform(X, y)

    assert X_bins["BILL_AMT1"].nunique() == 5
    assert X_bins["LIMIT_BAL"].nunique() == 5

    assert X_bins[X["LIMIT_BAL"] == 50000]["LIMIT_BAL"].unique() == np.array(-3)

    # Test that the labels are properly assigned. Because there are no specials in BILL_AMT1, there should be no extra
    # bins
    assert len(tbt.features_bucket_mapping_.get("BILL_AMT1").labels) == 6
    # check that the last label finishes with inf
    assert tbt.features_bucket_mapping_.get("BILL_AMT1").labels[0].startswith("[-inf")
    assert tbt.features_bucket_mapping_.get("BILL_AMT1").labels[4].endswith("inf)")

    # Test that the labels are properly assigned. Because there are 2 specials in LIMIT_BAL, there should be 2 extra
    # bins
    assert len(tbt.features_bucket_mapping_.get("LIMIT_BAL").labels) == 6
    # check that the labels match the specials dictionary
    assert (
        tbt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-3]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][0])
    )
    assert (
        tbt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-4]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][1])
    )

    # Assert a value error is raised if the specials contains features not defined in the bucketer.
    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}, "Undefinedfeature": {"1": [2]}}

    with pytest.raises(ValueError):
        DecisionTreeBucketer(
            variables=["LIMIT_BAL", "BILL_AMT1"], random_state=1, max_n_bins=3, specials=specials
        ).fit_transform(X, y)


def test_specials_equal_width_bucketer(df):
    """Test that when adding specials,the binner performs as expected.

    Context: special values should be binned in their own bin.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    ebt = EqualWidthBucketer(variables=["LIMIT_BAL", "BILL_AMT1"], n_bins=3, specials=specials)
    X_bins = ebt.fit_transform(X, y)

    assert X_bins["BILL_AMT1"].nunique() == 3
    assert X_bins["LIMIT_BAL"].nunique() == 5  # maximum n_bins +2 coming from the specials

    assert X_bins[X["LIMIT_BAL"] == 50000]["LIMIT_BAL"].unique() == np.array(-3)

    # Test that the labels are properly assigned. Because there are no specials in BILL_AMT1, there should be no extra
    # bins
    assert len(ebt.features_bucket_mapping_.get("BILL_AMT1").labels) == 4
    # check that the last label finishes with inf
    assert ebt.features_bucket_mapping_.get("BILL_AMT1").labels[0].startswith("(-inf")
    assert ebt.features_bucket_mapping_.get("BILL_AMT1").labels[2].endswith("inf]")

    # Test that the labels are properly assigned. Because there are 2 specials in LIMIT_BAL, there should be 2 extra
    # bins
    assert len(ebt.features_bucket_mapping_.get("LIMIT_BAL").labels) == 6
    # check that the labels match the specials dictionary
    assert (
        ebt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-3]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][0])
    )
    assert (
        ebt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-4]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][1])
    )

    # Assert a value error is raised if the specials contains features not defined in the bucketer.
    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}, "Undefinedfeature": {"1": [2]}}

    with pytest.raises(ValueError):
        EqualWidthBucketer(variables=["LIMIT_BAL", "BILL_AMT1"], n_bins=3, specials=specials).fit_transform(X, y)


def test_specials_equal_frequency_bucketer(df):
    """Test that when adding specials,the binner performs as expected.

    Context: special values should be binned in their own bin.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    ebt = EqualFrequencyBucketer(variables=["LIMIT_BAL", "BILL_AMT1"], n_bins=3, specials=specials)
    X_bins = ebt.fit_transform(X, y)

    assert X_bins["BILL_AMT1"].nunique() == 3
    assert X_bins["LIMIT_BAL"].nunique() == 5  # maximum n_bins +2 coming from the specials

    assert X_bins[X["LIMIT_BAL"] == 50000]["LIMIT_BAL"].unique() == np.array(-3)

    # Test that the labels are properly assigned. Because there are no specials in BILL_AMT1, there should be no extra
    # bins
    assert len(ebt.features_bucket_mapping_.get("BILL_AMT1").labels) == 4
    # check that the last label finishes with inf
    assert ebt.features_bucket_mapping_.get("BILL_AMT1").labels[0].startswith("(-inf")
    assert ebt.features_bucket_mapping_.get("BILL_AMT1").labels[2].endswith("inf]")

    # Test that the labels are properly assigned. Because there are 2 specials in LIMIT_BAL, there should be 2 extra
    # bins
    assert len(ebt.features_bucket_mapping_.get("LIMIT_BAL").labels) == 6
    # check that the labels match the specials dictionary
    assert (
        ebt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-3]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][0])
    )
    assert (
        ebt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-4]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][1])
    )

    # Assert a value error is raised if the specials contains features not defined in the bucketer.
    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}, "Undefinedfeature": {"1": [2]}}

    with pytest.raises(ValueError):
        EqualFrequencyBucketer(variables=["LIMIT_BAL", "BILL_AMT1"], n_bins=3, specials=specials).fit_transform(X, y)


def test_specials_optimal_bucketer(df):
    """Test that when adding specials,the binner performs as expected.

    Context: special values should be binned in their own bin.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    opt = OptimalBucketer(variables=["LIMIT_BAL"], max_n_bins=3, specials=specials)
    X_bins = opt.fit_transform(X, y)

    assert X_bins["LIMIT_BAL"].nunique() == 5

    assert X_bins[X["LIMIT_BAL"] == 50000]["LIMIT_BAL"].unique() == np.array(-3)

    # Test that the labels are properly assigned. Because there are 2 specials in LIMIT_BAL, there should be 2 extra
    # bins
    assert len(opt.features_bucket_mapping_.get("LIMIT_BAL").labels) == 6
    # check that the labels match the specials dictionary
    assert (
        opt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-3]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][0])
    )
    assert (
        opt.features_bucket_mapping_.get("LIMIT_BAL")
        .labels[-4]
        .endswith([key for key in specials["LIMIT_BAL"].keys()][1])
    )



import warnings
from collections import defaultdict

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from skorecard.bucketers.base_bucketer import BaseBucketer
from skorecard.metrics.metrics import woe_1d
from skorecard.utils.validation import ensure_dataframe


class WoeEncoder(BaseEstimator, TransformerMixin):
    """
    Transformer that encodes unique values in features to their Weight of Evidence estimation.

    **This class has been deprecated in favor of category_encoders.woe.WOEEncoder**

    Only works for binary classification (target y has 0 and 1 values).

    The weight of evidence is given by: `np.log( p(1) / p(0) )`
    The target probability ratio is given by: `p(1) / p(0)`

    For example in the variable colour, if the mean of the target = 1 for blue is 0.8 and
    the mean of the target = 0 is 0.2, blue will be replaced by: np.log(0.8/0.2) = 1.386
    if log_ratio is selected. Alternatively, blue will be replaced by 0.8 / 0.2 = 4 if ratio is selected.

    More formally:

    - for each unique value 𝑥,  consider the corresponding rows in the training set
    - compute what percentage of positives is in these rows, compared to the whole set
    - compute what percentage of negatives is in these rows, compared to the whole set
    - take the ratio of these percentages
    - take the natural logarithm of that ratio to get the weight of evidence corresponding to  𝑥,  so that  𝑊𝑂𝐸(𝑥)  is either positive or negative according to whether  𝑥  is more representative of positives or negatives

    More details:

    - [blogpost on weight of evidence](https://multithreaded.stitchfix.com/blog/2015/08/13/weight-of-evidence/)

    Example:

    ```python
    from skorecard import datasets
    from skorecard.preprocessing import WoeEncoder

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    we = WoeEncoder(variables=['EDUCATION'])
    we.fit_transform(X, y)
    we.fit_transform(X, y)['EDUCATION'].value_counts()
    ```

    Credits: Some inspiration taken from [feature_engine.categorical_encoders](https://feature-engine.readthedocs.io/en/latest/encoding/index.html).
    """  # noqa

    def __init__(self, epsilon=0.0001, variables=[], handle_unknown="value"):
        """
        Constructor for WoEEncoder.

        Args:
            epsilon (float): Amount to be added to relative counts in order to avoid division by zero in the WOE
                calculation.
            variables (list): The features to bucket. Uses all features if not defined.
            handle_unknown (str): How to handle any new values encountered in X on transform().
                options are 'return_nan', 'error' and 'value', defaults to 'value', which will assume WOE=0.
        """
        self.epsilon = epsilon
        self.variables = variables
        self.handle_unknown = handle_unknown

        warnings.warn(
            "This encoder will be deprecated. Please use category_encoders.woe.WOEEncoder instead.", DeprecationWarning
        )

    def fit(self, X, y):
        """Calculate the WOE for every column.

        Args:
            X (np.array): (binned) features
            y (np.array): target
        """
        assert self.epsilon >= 0
        # Check data
        X = ensure_dataframe(X)
        assert y is not None, "WoEBucketer needs a target y"
        y = BaseBucketer._check_y(y)

        y = y.astype(float)
        if len(np.unique(y)) > 2:
            raise AssertionError("WoEBucketer is only suited for binary classification")
        self.variables_ = BaseBucketer._check_variables(X, self.variables)

        # WoE currently does not support NAs
        # This is also flagged in self._more_tags()
        # We could treat missing values as a separate bin (-1) and thus handle seamlessly.
        BaseBucketer._check_contains_na(X, self.variables_)

        # scikit-learn requires checking that X has same shape on transform
        # this is because scikit-learn is still positional based (no column names used)
        self.n_train_features_ = X.shape[1]

        self.woe_mapping_ = {}
        for var in self.variables_:
            t = woe_1d(X[var], y, epsilon=self.epsilon)

            woe_dict = t["woe"].to_dict()
            # If new categories encountered, returns WoE = 0
            if self.handle_unknown == "value":
                woe_dict = defaultdict(int, woe_dict)

            self.woe_mapping_[var] = woe_dict

        return self

    def transform(self, X):
        """Transform X to weight of evidence encoding.

        Args:
            X (pd.DataFrame): dataset
        """
        assert self.handle_unknown in ["value", "error", "return_nan"]
        check_is_fitted(self)
        X = ensure_dataframe(X)

        if X.shape[1] != self.n_train_features_:
            msg = f"Number of features in X ({X.shape[1]}) is different "
            msg += f"from the number of features in X during fit ({self.n_train_features_})"
            raise ValueError(msg)

        for feature in self.variables_:
            woe_dict = self.woe_mapping_.get(feature)
            if self.handle_unknown == "error":
                new_cats = [x for x in list(X[feature].unique()) if x not in list(woe_dict.keys())]
                if len(new_cats) > 0:
                    msg = "WoEEncoder encountered unknown new categories "
                    msg += f"in column {feature} on .transform(): {new_cats}"
                    raise AssertionError(msg)

            X[feature] = X[feature].map(woe_dict)

        return X

    def _more_tags(self):
        """
        Estimator tags are annotations of estimators that allow programmatic inspection of their capabilities.

        See https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        """  # noqa
        return {"binary_only": True, "allow_nan": False}



import pytest
from sklearn.tree import DecisionTreeClassifier
import warnings

from skorecard.utils.validation import check_args


@pytest.mark.parametrize("args, nr_of_warnings", [({"hi": 1}, 1), ({"min_samples_leaf": 1}, 0)])
def test_checks_args(args, nr_of_warnings):
    """
    Tests checking arguments.
    """
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter("always")
        check_args(args, DecisionTreeClassifier)
        return len(caught_warnings) == nr_of_warnings



import numpy as np
import pytest
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted

from skorecard.bucketers import (
    AgglomerativeClusteringBucketer,
    AsIsCategoricalBucketer,
    AsIsNumericalBucketer,
    DecisionTreeBucketer,
    EqualFrequencyBucketer,
    EqualWidthBucketer,
    OptimalBucketer,
    OrdinalCategoricalBucketer,
)
from skorecard.bucketers.bucketers import UserInputBucketer
from skorecard.pipeline import BucketingProcess

BUCKETERS_WITH_SET_BINS = [
    EqualWidthBucketer,
    AgglomerativeClusteringBucketer,
    EqualFrequencyBucketer,
]

BUCKETERS_WITHOUT_SET_BINS = [
    OptimalBucketer,
    OrdinalCategoricalBucketer,
    AsIsNumericalBucketer,
    AsIsCategoricalBucketer,
    DecisionTreeBucketer,
]

# Except the very special UserInputBucketer of course :)
ALL_BUCKETERS = BUCKETERS_WITH_SET_BINS + BUCKETERS_WITHOUT_SET_BINS
ALL_BUCKETERS_WITH_BUCKETPROCESS = ALL_BUCKETERS + [BucketingProcess]


@pytest.mark.parametrize("bucketer", BUCKETERS_WITH_SET_BINS)
def test_single_bucket(bucketer, df) -> None:
    """Test that using n_bins=1 puts everything into 1 bucket."""
    BUCK = bucketer(n_bins=1, variables=["MARRIAGE"])
    x_t = BUCK.fit_transform(df)
    assert len(x_t["MARRIAGE"].unique()) == 1


@pytest.mark.parametrize("bucketer", BUCKETERS_WITH_SET_BINS)
def test_two_buckets(bucketer, df) -> None:
    """Test that using n_bins=1 puts everything into 2 buckets."""
    X = df
    y = df["default"].values

    BUCK = bucketer(n_bins=2, variables=["MARRIAGE"])
    BUCK.fit(X, y)
    x_t = BUCK.transform(X)
    assert len(x_t["MARRIAGE"].unique()) == 2


@pytest.mark.parametrize("bucketer", BUCKETERS_WITH_SET_BINS)
def test_three_bins(bucketer, df) -> None:
    """Test that we get the number of bins we request."""
    # Test single bin counts
    BUCK = bucketer(n_bins=3, variables=["MARRIAGE"])
    x_t = BUCK.fit_transform(df)
    assert len(x_t["MARRIAGE"].unique()) == 3


@pytest.mark.parametrize("bucketer", BUCKETERS_WITH_SET_BINS)
def test_error_input(bucketer):
    """Test that a non-int leads to problems in bins.

    Note input validation is done on fit, but before data validation.
    """
    with pytest.raises(AssertionError):
        bucketer(n_bins=[2]).fit(X=1, y=1)

    with pytest.raises(AssertionError):
        bucketer(n_bins=4.2, variables=["MARRIAGE"]).fit(X=1, y=1)


@pytest.mark.parametrize("bucketer", BUCKETERS_WITH_SET_BINS)
def test_missings_set(bucketer, df_with_missings) -> None:
    """Test all missing methods work for bucketers with set bins."""
    X = df_with_missings
    y = df_with_missings["default"].values

    BUCK = bucketer(n_bins=2, variables=["MARRIAGE"])
    BUCK.fit(X, y)
    X["MARRIAGE_trans"] = BUCK.transform(X)["MARRIAGE"]
    assert len(X["MARRIAGE_trans"].unique()) == 3
    assert X[np.isnan(X["MARRIAGE"])].shape[0] == X[X["MARRIAGE_trans"] == -1].shape[0]

    X = df_with_missings
    y = df_with_missings["default"].values

    BUCK = bucketer(n_bins=3, variables=["MARRIAGE", "LIMIT_BAL"], missing_treatment={"LIMIT_BAL": 1, "MARRIAGE": 0})
    BUCK.fit(X, y)
    X_trans = BUCK.transform(X)
    assert len(X_trans["MARRIAGE"].unique()) == 3
    assert len(X_trans["LIMIT_BAL"].unique()) == 3

    X["MARRIAGE_TRANS"] = X_trans["MARRIAGE"]
    assert X[np.isnan(X["MARRIAGE"])]["MARRIAGE_TRANS"].sum() == 0  # Sums to 0 as they are all in bucket 0

    assert "| Missing" in [f for f in BUCK.features_bucket_mapping_.get("MARRIAGE").labels.values()][0]
    assert "| Missing" in [f for f in BUCK.features_bucket_mapping_.get("LIMIT_BAL").labels.values()][1]

    BUCK = bucketer(n_bins=3, variables=["MARRIAGE", "LIMIT_BAL"], missing_treatment="most_frequent")
    BUCK.fit(X, y)

    for feature in ["MARRIAGE", "LIMIT_BAL"]:
        assert (
            "Missing"
            in BUCK.bucket_table(feature).sort_values("Count", ascending=False).reset_index(drop=True)["label"][0]
        )

    BUCK_risk = bucketer(n_bins=3, variables=["MARRIAGE", "EDUCATION"], missing_treatment="most_risky")
    BUCK_risk.fit(X, y)

    BUCK_norisk = bucketer(n_bins=3, variables=["MARRIAGE", "EDUCATION"])
    BUCK_norisk.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the riskiest bucket when missings are in a separate bucket
        riskiest_bucket = (
            BUCK_norisk.bucket_table(feature)[BUCK_norisk.bucket_table(feature)["bucket"] >= 0]
            .sort_values("Event Rate", ascending=False)
            .reset_index(drop=True)["bucket"][0]
        )
        assert (
            "Missing"
            in BUCK_risk.bucket_table(feature)[
                BUCK_risk.bucket_table(feature)["bucket"] == riskiest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_risk = bucketer(n_bins=3, variables=["MARRIAGE", "EDUCATION"], missing_treatment="least_risky")
    BUCK_risk.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the safest bucket when missings are in a separate bucket
        safest_bucket = (
            BUCK_norisk.bucket_table(feature)[BUCK_norisk.bucket_table(feature)["bucket"] >= 0]
            .sort_values("Event Rate", ascending=True)
            .reset_index(drop=True)["bucket"][0]
        )
        assert (
            "Missing"
            in BUCK_risk.bucket_table(feature)[
                BUCK_risk.bucket_table(feature)["bucket"] == safest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_neutral = bucketer(n_bins=3, variables=["MARRIAGE", "EDUCATION"], missing_treatment="neutral")
    BUCK_neutral.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the bucket with WoE closest to 0
        table = BUCK_norisk.bucket_table(feature)
        table["WoE"] = np.abs(table["WoE"])
        closest_bucket = table[table["Count"] > 0].sort_values("WoE").reset_index(drop=True)["bucket"][0]
        assert (
            "Missing"
            in BUCK_neutral.bucket_table(feature)[
                BUCK_neutral.bucket_table(feature)["bucket"] == closest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_similar = bucketer(n_bins=3, variables=["MARRIAGE", "EDUCATION"], missing_treatment="similar")
    BUCK_similar.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the bucket with WoE closest to 0
        table = BUCK_norisk.bucket_table(feature)
        missing_WoE = table[table["label"] == "Missing"]["WoE"].values[0]
        table["New_WoE"] = np.abs(table["WoE"] - missing_WoE)
        closest_bucket = table[table["label"] != "Missing"].sort_values("New_WoE").reset_index(drop=True)["bucket"][0]
        assert (
            "Missing"
            in BUCK_similar.bucket_table(feature)[
                BUCK_similar.bucket_table(feature)["bucket"] == closest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_passthrough = bucketer(n_bins=3, variables=["MARRIAGE", "EDUCATION"], missing_treatment="passthrough")
    BUCK_passthrough.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the bucket id with 'Missing' label
        table = BUCK_passthrough.bucket_table(feature)
        impute_missing = "Soup37120"
        table = table.fillna(impute_missing)
        missing_bucket = table[table["label"] == "Missing"]["bucket"].values[0]
        assert missing_bucket == impute_missing
        X_trans = BUCK_passthrough.transform(X)[[feature]]
        original_index_missings = X[X[feature].isnull()][feature].index
        trans_index_missings = X_trans[X_trans[feature].isnull()][feature].index
        assert all(original_index_missings[i] == trans_index_missings[i] for i in range(len(original_index_missings)))


@pytest.mark.parametrize("bucketer", BUCKETERS_WITHOUT_SET_BINS)
def test_missings_without_set(bucketer, df_with_missings) -> None:
    """Test all missing methods work for bucketers without set bins."""
    X = df_with_missings
    y = df_with_missings["default"].values

    BUCK = bucketer(variables=["MARRIAGE", "EDUCATION"], missing_treatment="most_frequent")
    BUCK.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        assert (
            "Missing"
            in BUCK.bucket_table(feature).sort_values("Count", ascending=False).reset_index(drop=True)["label"][0]
        )

    BUCK_risk = bucketer(variables=["MARRIAGE", "EDUCATION"], missing_treatment="most_risky")
    BUCK_risk.fit(X, y)

    BUCK_norisk = bucketer(variables=["MARRIAGE", "EDUCATION"])
    BUCK_norisk.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the riskiest bucket when missings are in a separate bucket
        riskiest_bucket = (
            BUCK_norisk.bucket_table(feature)[BUCK_norisk.bucket_table(feature)["bucket"] >= 0]
            .sort_values("Event Rate", ascending=False)
            .reset_index(drop=True)["bucket"][0]
        )
        assert (
            "Missing"
            in BUCK_risk.bucket_table(feature)[
                BUCK_risk.bucket_table(feature)["bucket"] == riskiest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_risk = bucketer(variables=["MARRIAGE", "EDUCATION"], missing_treatment="least_risky")
    BUCK_risk.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the safest bucket when missings are in a separate bucket
        safest_bucket = (
            BUCK_norisk.bucket_table(feature)[BUCK_norisk.bucket_table(feature)["bucket"] >= 0]
            .sort_values("Event Rate", ascending=True)
            .reset_index(drop=True)["bucket"][0]
        )
        assert (
            "Missing"
            in BUCK_risk.bucket_table(feature)[
                BUCK_risk.bucket_table(feature)["bucket"] == safest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_neutral = bucketer(variables=["MARRIAGE", "EDUCATION"], missing_treatment="neutral")
    BUCK_neutral.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the bucket with WoE closest to 0
        table = BUCK_norisk.bucket_table(feature)
        table["WoE"] = np.abs(table["WoE"])
        closest_bucket = table[table["Count"] > 0].sort_values("WoE").reset_index(drop=True)["bucket"][0]
        assert (
            "Missing"
            in BUCK_neutral.bucket_table(feature)[
                BUCK_neutral.bucket_table(feature)["bucket"] == closest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_similar = bucketer(variables=["MARRIAGE", "EDUCATION"], missing_treatment="similar")
    BUCK_similar.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the bucket with WoE closest to 0
        table = BUCK_norisk.bucket_table(feature)
        missing_WoE = table[table["label"] == "Missing"]["WoE"].values[0]
        table["New_WoE"] = np.abs(table["WoE"] - missing_WoE)
        closest_bucket = table[table["label"] != "Missing"].sort_values("New_WoE").reset_index(drop=True)["bucket"][0]
        assert (
            "Missing"
            in BUCK_similar.bucket_table(feature)[
                BUCK_similar.bucket_table(feature)["bucket"] == closest_bucket
            ].reset_index()["label"][0]
        )

    BUCK_passthrough = bucketer(variables=["MARRIAGE", "EDUCATION"], missing_treatment="passthrough")
    BUCK_passthrough.fit(X, y)

    for feature in ["MARRIAGE", "EDUCATION"]:
        # look at the bucket id with 'Missing' label
        table = BUCK_passthrough.bucket_table(feature)
        impute_missing = "Soup37120"
        table = table.fillna(impute_missing)
        missing_bucket = table[table["label"] == "Missing"]["bucket"].values[0]
        assert missing_bucket == impute_missing
        X_trans = BUCK_passthrough.transform(X)[[feature]]
        original_index_missings = X[X[feature].isnull()][feature].index
        trans_index_missings = X_trans[X_trans[feature].isnull()][feature].index
        assert all(original_index_missings[i] == trans_index_missings[i] for i in range(len(original_index_missings)))


@pytest.mark.parametrize("bucketer", ALL_BUCKETERS)
def test_type_error_input(bucketer, df):
    """Test that input is always a dataFrame."""
    y = df["default"].values
    X = df.drop(columns=["pet_ownership", "default"])
    pipe = make_pipeline(
        StandardScaler(),
        bucketer(variables=["BILL_AMT1"]),
    )
    with pytest.raises(AssertionError):
        pipe.fit_transform(X, y)


@pytest.mark.parametrize("bucketer", ALL_BUCKETERS)
def test_is_not_fitted(bucketer):
    """
    Make sure we didn't make any mistakes when building a bucketer.
    """
    BUCK = bucketer()
    with pytest.raises(NotFittedError):
        check_is_fitted(BUCK)


@pytest.mark.parametrize("bucketer", ALL_BUCKETERS)
def test_ui_bucketer(bucketer, df):
    """
    Make sure we didn't make any mistakes when building a bucketer.
    """
    BUCK = bucketer()
    # we drop BILL_AMT1 because that one needs prebucketing for some bucketers.
    df = df.drop(columns=["pet_ownership", "BILL_AMT1"])
    X = df
    y = df["default"].values
    X_trans = BUCK.fit_transform(X, y)

    uib = UserInputBucketer(BUCK.features_bucket_mapping_)
    assert X_trans.equals(uib.transform(X))


@pytest.mark.parametrize("bucketer", ALL_BUCKETERS)
def test_zero_indexed(bucketer, df):
    """Test that bins are zero-indexed.

    When no missing values are present, no specials defined,
    bucket transforms should be zero indexed.

    Note that -2 (for 'other') is also allowed,
    f.e. OrdinalCategoricalBucketer puts less frequents cats there.
    """
    BUCK = bucketer()

    y = df["default"].values
    # we drop BILL_AMT1 because that one needs prebucketing for some bucketers.
    x_t = BUCK.fit_transform(df.drop(columns=["pet_ownership", "BILL_AMT1"]), y)
    assert x_t["MARRIAGE"].min() in [0, -2]
    assert x_t["EDUCATION"].min() in [0, -2]
    assert x_t["LIMIT_BAL"].min() in [0, -2]


@pytest.mark.parametrize("bucketer", ALL_BUCKETERS)
def test_remainder_argument_no_bins(bucketer, df):
    """Test remainder argument works."""
    BUCK = bucketer(variables=["LIMIT_BAL"], remainder="drop")
    X = df
    y = df["default"].values

    BUCK.fit(X, y)
    X_trans = BUCK.transform(X)
    assert X_trans.columns == "LIMIT_BAL"

    BUCK = bucketer(variables=["LIMIT_BAL"], remainder="passthrough")
    BUCK.fit(X, y)
    X_trans = BUCK.transform(X)
    assert set(X_trans.columns) == set(X.columns)


@pytest.mark.parametrize("bucketer", ALL_BUCKETERS)
def test_summary_no_bins(bucketer, df):
    """Test summary works."""
    BUCK = bucketer(variables=["LIMIT_BAL"], remainder="passthrough")
    X = df
    y = df["default"].values
    BUCK.fit(X, y)
    summary_table = BUCK.summary()
    assert summary_table.shape[0] == 6
    assert set(summary_table.columns) == {"column", "num_prebuckets", "num_buckets", "IV_score", "dtype"}



import pathlib
import warnings
from copy import deepcopy
from typing import Dict, List, Optional, TypeVar

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import make_pipeline
from sklearn.utils.validation import check_is_fitted

from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer
from skorecard.features_bucket_mapping import FeaturesBucketMapping, merge_features_bucket_mapping
from skorecard.pipeline import to_skorecard_pipeline
from skorecard.pipeline.pipeline import _get_all_steps
from skorecard.reporting import build_bucket_table
from skorecard.reporting.plotting import PlotBucketMethod, PlotPreBucketMethod
from skorecard.reporting.report import BucketTableMethod, SummaryMethod
from skorecard.utils import NotBucketedError, NotPreBucketedError
from skorecard.utils.exceptions import NotInstalledError
from skorecard.utils.validation import ensure_dataframe, is_fitted

# JupyterDash
try:
    from jupyter_dash import JupyterDash
except ModuleNotFoundError:
    JupyterDash = NotInstalledError("jupyter-dash", "dashboard")

try:
    import dash_bootstrap_components as dbc
except ModuleNotFoundError:
    dbc = NotInstalledError("dash_bootstrap_components", "dashboard")


from skorecard.apps.app_callbacks import add_bucketing_process_callbacks
from skorecard.apps.app_layout import add_bucketing_process_layout

PathLike = TypeVar("PathLike", str, pathlib.Path)


class BucketingProcess(
    BaseEstimator,
    TransformerMixin,
    BucketTableMethod,
    PlotBucketMethod,
    PlotPreBucketMethod,
    SummaryMethod,
):
    """
    A two-step bucketing pipeline allowing for pre-bucketing before bucketing.

    Often you want to pre-bucket features (f.e. to 100 buckets) before bucketing to a smaller set.
    This brings some additional challenges around propagating specials and defining a bucketer that is able to go from raw data to final bucket.
    This class facilicates the process and also provides all regular methods and attributes:

    - `.summary()`: See which columns are bucketed
    - `.plot_bucket()`: Plot buckets of a column
    - `.bucket_table()`: Table with buckets of a column
    - `.save_to_yaml()`: Save information necessary for bucketing to a YAML file
    - `.features_bucket_mapping_`: Access bucketing information

    Example:

    ```python
    from skorecard import datasets
    from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer, AsIsCategoricalBucketer
    from skorecard.pipeline import BucketingProcess
    from sklearn.pipeline import make_pipeline

    df = datasets.load_uci_credit_card(as_frame=True)
    y = df["default"]
    X = df.drop(columns=["default"])

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucketing_process = BucketingProcess(
        specials={'LIMIT_BAL': {'=400000.0' : [400000.0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type='categorical', max_n_bins=10, min_bin_size=0.05),
        )
    )

    bucketing_process.fit(X, y)

    # Details
    bucketing_process.summary() # all vars, and # buckets
    bucketing_process.bucket_table("LIMIT_BAL")
    bucketing_process.plot_bucket("LIMIT_BAL")
    bucketing_process.prebucket_table("LIMIT_BAL")
    bucketing_process.plot_prebucket("LIMIT_BAL")
    ```
    """  # noqa

    def __init__(
        self,
        prebucketing_pipeline=make_pipeline(DecisionTreeBucketer(max_n_bins=50, min_bin_size=0.02)),
        bucketing_pipeline=make_pipeline(OptimalBucketer(max_n_bins=6, min_bin_size=0.05)),
        variables: List = [],
        specials: Dict = {},
        random_state: Optional[int] = None,
        remainder="passthrough",
    ):
        """
        Define a BucketingProcess to first prebucket and then bucket multiple columns in one go.

        Args:
            prebucketing_pipeline (Pipeline): The scikit-learn pipeline that does pre-bucketing.
                Defaults to an all-numeric DecisionTreeBucketer pipeline.
            bucketing_pipeline (Pipeline): The scikit-learn pipeline that does bucketing.
                Defaults to an all-numeric OptimalBucketer pipeline.
                Must transform same features as the prebucketing pipeline.
            variables (list): The features to bucket. Uses all features if not defined.
            specials: (nested) dictionary of special values that require their own binning.
                Will merge when specials are also defined in any bucketers in a (pre)bucketing pipeline, and overwrite in case there are shared keys.
                The dictionary has the following format:
                 {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                For every feature that needs a special value, a dictionary must be passed as value.
                This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                in that bucket.
                When special values are defined, they are not considered in the fitting procedure.
            remainder (str): How we want the non-specified columns to be transformed. It must be in ["passthrough", "drop"].
                passthrough (Default): all columns that were not specified in "variables" will be passed through.
                drop: all remaining columns that were not specified in "variables" will be dropped.
        """  # noqa
        # Save original input params
        # We overwrite the input later, so we need to save
        # original so we can clone instances
        # https://scikit-learn.org/dev/developers/develop.html#cloning
        # https://scikit-learn.org/dev/developers/develop.html#get-params-and-set-params
        # Assigning the variable in the init to the attribute with the same name is a requirement of
        # sklearn.base.BaseEstimator. See the notes in
        # https://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html#sklearn.base.BaseEstimator
        self.prebucketing_pipeline = prebucketing_pipeline
        self.bucketing_pipeline = bucketing_pipeline
        self.remainder = remainder
        self.variables = variables
        self.specials = specials
        self.random_state = random_state

    @property
    def name(self):
        """
        To be able to identity the bucketingprocess in a pipeline.
        """
        return "bucketingprocess"

    def fit(self, X, y=None):
        """
        Fit the prebucketing and bucketing pipeline with `X`, `y`.

        Args:
            X (pd.DataFrame): Data to fit on.
            y (np.array, optional): target. Defaults to None.
        """
        X = ensure_dataframe(X)

        # input validation
        assert self.remainder in ["passthrough", "drop"]

        # Convert to skorecard pipelines
        # This does some checks on the pipelines
        # and adds some convenience methods to the pipeline.
        self.pre_pipeline_ = to_skorecard_pipeline(deepcopy(self.prebucketing_pipeline))
        self.pipeline_ = to_skorecard_pipeline(deepcopy(self.bucketing_pipeline))

        # Add/Overwrite specials to all pre-bucketers
        for step in _get_all_steps(self.pre_pipeline_):
            if hasattr(step, "specials") and len(step.specials) != 0 and len(self.specials) != 0:
                # note, specials defined BucketingProcess level
                # will overwrite any specials on bucketer level.
                warnings.warn(f"Overwriting specials of {step} with specials of bucketingprocess", UserWarning)
                step.specials = {**step.specials, **self.specials}
            else:
                step.specials = self.specials

            if len(self.variables) != 0:
                if len(step.variables) != 0:
                    warnings.warn(f"Overwriting variables of {step} with variables of bucketingprocess", UserWarning)
                step.variables = self.variables

            # Overwrite random_state to bucketers
            if hasattr(step, "random_state") and self.random_state is not None:
                if step.random_state is not None:
                    warnings.warn(
                        f"Overwriting random_state of {step} with random_state of bucketingprocess", UserWarning
                    )
                step.random_state = self.random_state

        # Overwrite variables to all bucketers
        if len(self.variables) != 0:
            for step in _get_all_steps(self.pipeline_):
                if len(step.variables) != 0:
                    warnings.warn(f"Overwriting variables of {step} with variables of bucketingprocess", UserWarning)
                step.variables = self.variables

        # Overwrite random_state to bucketers
        for step in _get_all_steps(self.pipeline_):
            if hasattr(step, "random_state") and self.random_state is not None:
                if step.random_state is not None:
                    warnings.warn(
                        f"Overwriting random_state of {step} with random_state of bucketingprocess", UserWarning
                    )
                step.random_state = self.random_state

        self._prebucketing_specials = self.specials
        self._bucketing_specials = dict()  # will be determined later.

        # Fit the prebucketing pipeline
        X_prebucketed_ = self.pre_pipeline_.fit_transform(X, y)
        assert isinstance(X_prebucketed_, pd.DataFrame)

        # Calculate the prebucket tables.
        self.prebucket_tables_ = dict()
        for column in X.columns:
            if column in self.pre_pipeline_.features_bucket_mapping_.maps.keys():
                self.prebucket_tables_[column] = build_bucket_table(
                    X, y, column=column, bucket_mapping=self.pre_pipeline_.features_bucket_mapping_.get(column)
                )

        # Find the new bucket numbers of the specials after prebucketing,
        for var, var_specials in self._prebucketing_specials.items():
            bucket_labels = self.pre_pipeline_.features_bucket_mapping_.get(var).labels
            new_specials = _find_remapped_specials(bucket_labels, var_specials)
            if len(new_specials):
                self._bucketing_specials[var] = new_specials

        # Then assign the new specials to all bucketers in the bucketing pipeline
        for step in self.pipeline_.steps:
            if not isinstance(step, tuple):
                step.specials = self._bucketing_specials
            else:
                step[1].specials = self._bucketing_specials

        # Fit the bucketing pipeline
        # And save the bucket mapping
        self.pipeline_.fit(X_prebucketed_, y)

        # Make sure all columns that are bucketed have also been pre-bucketed.
        not_prebucketed = []
        for col in self.pipeline_.features_bucket_mapping_.columns:
            if self.pipeline_.features_bucket_mapping_.get(col).type == "numerical":
                if col not in self.pre_pipeline_.features_bucket_mapping_.columns:
                    not_prebucketed.append(col)
        if len(not_prebucketed):
            msg = "These numerical columns are bucketed but have not been pre-bucketed: "
            msg += f"{', '.join(not_prebucketed)}.\n"
            msg += "Consider adding a numerical bucketer to the prebucketing pipeline,"
            msg += "for example AsIsNumericalBucketer or DecisionTreeBucketer."
            raise NotPreBucketedError(msg)

        # Make sure all columns that have been pre-bucketed also have been bucketed
        not_bucketed = []
        for col in self.pre_pipeline_.features_bucket_mapping_.columns:
            if self.pre_pipeline_.features_bucket_mapping_.get(col).type == "numerical":
                if col not in self.pipeline_.features_bucket_mapping_.columns:
                    not_bucketed.append(col)
        if len(not_bucketed):
            msg = "These numerical columns are prebucketed but have not been bucketed: "
            msg += f"{', '.join(not_bucketed)}.\n"
            msg += "Consider updating the bucketing pipeline."
            raise NotBucketedError(msg)

        # calculate the bucket tables.
        self.bucket_tables_ = dict()
        for column in X.columns:
            if column in self.pipeline_.features_bucket_mapping_.maps.keys():
                self.bucket_tables_[column] = build_bucket_table(
                    X_prebucketed_,
                    y,
                    column=column,
                    bucket_mapping=self.pipeline_.features_bucket_mapping_.get(column),
                )

        # Calculate the summary
        self._generate_summary(X, y)

        return self

    def fit_interactive(self, X, y=None, mode="external", **server_kwargs):
        """
        Fit a bucketer and then interactive edit the fit using a dash app.

        Note we are using a [jupyterdash](https://medium.com/plotly/introducing-jupyterdash-811f1f57c02e) app,
        which supports 3 different modes:

        - 'external' (default): Start dash server and print URL
        - 'inline': Start dash app inside an Iframe in the jupyter notebook
        - 'jupyterlab': Start dash app as a new tab inside jupyterlab

        """
        # We need to make sure we only fit if not already fitted
        # This prevents a user losing manually defined boundaries
        # when re-running .fit_interactive()
        if not is_fitted(self):
            self.fit(X, y)

        self.app = JupyterDash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        add_bucketing_process_layout(self)
        add_bucketing_process_callbacks(self, X, y)
        self.app.run_server(mode=mode, **server_kwargs)

    def transform(self, X):
        """
        Transform `X` through the prebucketing and bucketing pipelines.
        """
        check_is_fitted(self)
        X_prebucketed = self.pre_pipeline_.transform(X)

        new_X = self.pipeline_.transform(X_prebucketed)

        if self.remainder == "drop":
            return new_X[self.variables]
        else:
            return new_X

    def save_yml(self, fout: PathLike) -> None:
        """
        Save the features bucket to a yaml file.

        Args:
            fout: path for output file
        """
        check_is_fitted(self)
        fbm = self.features_bucket_mapping_
        if isinstance(fbm, dict):
            FeaturesBucketMapping(fbm).save_yml(fout)
        else:
            fbm.save_yml(fout)

    @property
    def features_bucket_mapping_(self):
        """
        Returns a `FeaturesBucketMapping` instance.

        In normal bucketers, you can access `.features_bucket_mapping_`
        to retrieve a `FeaturesBucketMapping` instance. This contains
        all the info you need to transform values into their buckets.

        In this class, we basically have a two step bucketing process:
        first prebucketing, and then we bucket the prebuckets.

        In order to still be able to use BucketingProcess as if it were a normal bucketer,
        we'll need to merge both into one.
        """
        check_is_fitted(self)

        return merge_features_bucket_mapping(
            self.pre_pipeline_.features_bucket_mapping_, self.pipeline_.features_bucket_mapping_
        )

    def prebucket_table(self, column: str) -> pd.DataFrame:
        """
        Generates the statistics for the buckets of a particular column.

        An example is seen below:

        pre-bucket | label      | Count | Count (%) | Non-event | Event | Event Rate | WoE   | IV   | bucket
        -----------|------------|-------|-----------|-----------|-------|------------|-------|------|------
        0          | (-inf, 1.0)| 479   | 7.98      | 300       | 179   |  37.37     |  0.73 | 0.05 | 0
        1          | [1.0, 2.0) | 370   | 6.17      | 233       | 137   |  37.03     |  0.71 | 0.04 | 0

        Args:
            column (str): The column we wish to analyse

        Returns:
            df (pd.DataFrame): A pandas dataframe of the format above
        """  # noqa
        check_is_fitted(self)
        if column not in self.prebucket_tables_.keys():
            raise ValueError(f"column '{column}' was not part of the pre-bucketing process")

        table = self.prebucket_tables_.get(column)
        table = table.rename(columns={"bucket_id": "pre-bucket"})

        # Find bucket for each pre-bucket
        bucket_mapping = self.pipeline_.features_bucket_mapping_.get(column)
        table["bucket"] = bucket_mapping.transform(table["pre-bucket"])

        # Find out missing bucket
        if -1 in table["pre-bucket"].values:
            table.loc[table["pre-bucket"] == -1, "bucket"] = bucket_mapping.transform([np.nan])[0]

        # Find out the 'other' bucket
        if bucket_mapping.type == "categorical" and -2 in table["pre-bucket"].values:
            something_random = "84a088e251d2fa058f37145222e536dc"
            table.loc[table["pre-bucket"] == -2, "bucket"] = bucket_mapping.transform([something_random])[0]

        return table

    def _more_tags(self):
        """
        Estimator tags are annotations of estimators that allow programmatic inspection of their capabilities.

        See https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        """  # noqa
        return {"binary_only": True}


def _find_remapped_specials(bucket_labels: Dict, var_specials: Dict) -> Dict:
    """
    Remaps the specials after the prebucketing process.

    Basically, every bucketer in the bucketing pipeline will now need to
    use the prebucketing bucket as a different special value,
    because prebucketing put the specials into a bucket.

    Args:
        bucket_labels (dict): The label for each unique bucket of a variable
        var_specials (dict): The specials for a variable, if any.
    """
    if bucket_labels is None or var_specials is None:
        return {}

    new_specials = {}
    for label in var_specials.keys():
        for bucket, bucket_label in bucket_labels.items():
            if bucket_label == f"Special: {label}":
                new_specials[label] = [bucket]

    return new_specials



import numpy as np
import pytest
from optbinning import OptimalBinning
from sklearn.pipeline import make_pipeline

from skorecard.bucket_mapping import BucketMapping
from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer
from skorecard.utils import NotPreBucketedError


def test_optimal_binning_prebinning(df):
    """Ensure we have prevented prebinning correctly.

    optbinning.OptimalBinning() does pre-binning by default.
    In skorecard, we want more control, so force the user to do this explicitly.

    Extracting only the optimizer from optbinning is quite involved.
    Instead, we'll let the prebinning (a DecisionTreeClassifier) use our user defined splits.
    These splits are simply the unique values from the prebucketed feature.

    This tests checks the output is not changed.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"].values

    # optbinning saves the user_defined_splits for prebinning
    # https://github.com/guillermo-navas-palencia/optbinning/blob/396b9bed97581094167c9eb4744c2fd1fb5c7408/optbinning/binning/binning.py#L741
    # then applies np.digitize(right=False)
    # https://github.com/guillermo-navas-palencia/optbinning/blob/396b9bed97581094167c9eb4744c2fd1fb5c7408/optbinning/binning/binning.py#L1048
    # If these are the prebuckted values
    x = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    # Internally this happens:
    binned_x = np.digitize(x=x, bins=x, right=False)
    # which means a shift by 1
    assert all(binned_x == x + 1)
    # which means internally it is shifted by 1, but that doesn't really matter,
    # as tests below show, because still same unique buckets.

    # transform a feature using OptBinning's prebucket + optimal binning
    binner = OptimalBinning(
        name="BILL_AMT1",
        dtype="numerical",
        solver="cp",
        monotonic_trend="auto_asc_desc",
        min_prebin_size=0.05,
        max_n_prebins=20,
        min_bin_size=0.05,
        max_n_bins=10,
        cat_cutoff=0.05,
        time_limit=25,
    )
    old = binner.fit_transform(X["BILL_AMT1"], y, metric="indices")

    # transform a feature using skorecard prebinning + optimal binning
    pipe = make_pipeline(
        DecisionTreeBucketer(variables=["BILL_AMT1"], max_n_bins=20, min_bin_size=0.05),
        OptimalBucketer(variables=["BILL_AMT1"], max_n_bins=10, min_bin_size=0.05),
    )
    new = pipe.fit_transform(X, y)["BILL_AMT1"]

    assert all(np.equal(old, new.values))

    # now for categoricals


def test_optimal_binning_numerical(df):
    """Tests the wrapper of optbinning.OptimalBinning()."""
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"].values

    X_prebucketed = DecisionTreeBucketer(max_n_bins=20, min_bin_size=0.05).fit_transform(X, y)

    obt = OptimalBucketer(variables=["LIMIT_BAL", "BILL_AMT1"])
    with pytest.raises(NotPreBucketedError):
        obt.fit(X, y)
    obt.fit(X_prebucketed, y)
    X_trans = obt.transform(X_prebucketed)
    assert len(X_trans["LIMIT_BAL"].unique()) == 9
    assert len(X_trans["BILL_AMT1"].unique()) == 6

    obt = OptimalBucketer(variables=["LIMIT_BAL"])
    obt.fit(X_prebucketed, y)
    X_trans = obt.transform(X_prebucketed)
    assert len(X_trans["LIMIT_BAL"].unique()) == 9

    # Note our bins were 1-indexed in the past
    # This unit test is there to make sure our bins are zero-indexed
    obt.fit(X, y)
    assert all(obt.features_bucket_mapping_.get("LIMIT_BAL").transform([0, 30_000]) == np.array([0, 1]))

    # optb.binning_table.build() # TODO Cleanup
    # optb.splits
    # obt.features_bucket_mapping_.get('LIMIT_BAL')


def test_optimal_binning_categorical(df):
    """Test categoricals."""
    X = df[["LIMIT_BAL", "BILL_AMT1", "EDUCATION"]]
    y = df["default"].values

    obt = OptimalBucketer(variables=["EDUCATION"], variables_type="categorical")
    obt.fit(X, y)
    X_trans = obt.transform(X)
    assert len(X_trans["EDUCATION"].unique()) == 3
    assert obt.features_bucket_mapping_.get("EDUCATION") == BucketMapping(
        feature_name="EDUCATION",
        type="categorical",
        map={0: 0, 6: 0, 4: 0, 5: 0, 1: 0, 3: 1, 2: 2},
        right=False,
    )

    optb = OptimalBinning(
        name="EDUCATION", dtype="categorical", solver="cp", cat_cutoff=None, max_n_prebins=100, max_n_bins=10
    )
    optb.fit(X["EDUCATION"], y)
    ref = optb.transform(X["EDUCATION"], metric="indices")

    assert all(X_trans["EDUCATION"].values == ref)


def test_raises_prebucketing_error(df):
    """
    Test prebucketing error.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1", "EDUCATION"]]
    y = df["default"].values

    obt = OptimalBucketer(variables=["BILL_AMT1"])
    with pytest.raises(NotPreBucketedError):
        obt.fit_transform(X, y)


def test_optbinning_with_specials(df):
    """
    Test adding specials.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1", "EDUCATION"]]
    y = df["default"].values

    specials = {"LIMIT_BAL": {"=50000": [50000], "in [20001,30000]": [20000, 30000]}}

    obt = OptimalBucketer(variables=["LIMIT_BAL"], specials=specials)
    obt.fit_transform(X, y)


def test_optimal_binning_categorical_specials(df):
    """
    Test categoricals with specials.
    """
    # WIP - currently not implemented yet # TODO Cleanup
    X = df[["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "pet_ownership"]]
    y = df["default"].values

    specials = {"EDUCATION": {"special_one": [0]}}
    obt = OptimalBucketer(variables=["EDUCATION"], variables_type="categorical", specials=specials)
    obt.fit(X, y)
    X_trans = obt.transform(X)
    assert len(X_trans["EDUCATION"].unique()) == 4

    assert obt.transform(X)["EDUCATION"][X["EDUCATION"] == 0].shape[0] == 1

    X = df[["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "pet_ownership"]]
    y = df["default"].values
    specials = {"pet_ownership": {"indoor pets outside water": ["dog lover", "cat lover"]}}
    obt = OptimalBucketer(variables=["pet_ownership"], max_n_bins=2, variables_type="categorical", specials=specials)
    X_trans = obt.fit_transform(X, y)
    # import pdb; pdb.set_trace(); # TODO Cleanup
    assert len(X_trans.pet_ownership.unique()) == 2 + 1  # 2 bins plus 1 specials


def test_optimal_bucketing_cats(df):
    """
    Test with categorical strings.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "pet_ownership"]]
    y = df["default"].values
    # specials = {'pet_ownership': {"indoor pets outside water": ['dog lover','cat lover']}}

    # obt.features_bucket_mapping_.get('pet_ownership') # TODO Cleanup
    obt = OptimalBucketer(variables=["pet_ownership"], variables_type="categorical")
    X_trans = obt.fit_transform(X, y)
    # assert len(X_trans.pet_ownership.unique()) == 5

    obt = OptimalBucketer(variables=["pet_ownership"], max_n_bins=3, variables_type="categorical")
    X_trans = obt.fit_transform(X, y)
    # assert len(X_trans.pet_ownership.unique()) == 3+1

    obt = OptimalBucketer(variables=["pet_ownership"], max_n_bins=2, variables_type="categorical")
    X_trans = obt.fit_transform(X, y)
    # assert len(X_trans.pet_ownership.unique()) == 2+1

    obt = OptimalBucketer(variables=["LIMIT_BAL"], max_n_bins=2)
    X_trans = obt.fit_transform(X, y)
    assert len(X_trans["LIMIT_BAL"].unique()) == 2


def test_missing_categorical_manual(df_with_missings):
    """
    Test.
    """
    X = df_with_missings[["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "pet_ownership"]]
    y = df_with_missings["default"].values
    obt = OptimalBucketer(
        variables=["pet_ownership", "EDUCATION"],
        max_n_bins=2,
        variables_type="categorical",
        missing_treatment={"pet_ownership": 0},
    )
    obt.fit(X[["pet_ownership", "EDUCATION"]], y)
    X_trans = obt.transform(X[["pet_ownership", "EDUCATION"]])  # ['pet_ownership']
    X["pet_ownership_trans"] = X_trans["pet_ownership"]
    X["default"] = y
    assert (
        X[X["pet_ownership"].isnull()]["pet_ownership_trans"].sum() == 0
    )  # Sums to 0 because all missings in bucket 0


def test_missing_numerical_manual(df_with_missings):
    """
    Test.
    """
    X = df_with_missings[["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "pet_ownership"]]
    y = df_with_missings["default"].values
    obt = OptimalBucketer(
        variables=["LIMIT_BAL"], max_n_bins=2, variables_type="numerical", missing_treatment={"LIMIT_BAL": 0}
    )
    obt.fit(X[["LIMIT_BAL"]], y)
    X["LIMIT_BAL_trans"] = obt.transform(X[["LIMIT_BAL"]])
    assert X[np.isnan(X["LIMIT_BAL"])]["LIMIT_BAL_trans"].sum() == 0  # Sums to 0 because all missings in bucket 0



import pytest
from sklearn.exceptions import NotFittedError

from skorecard.bucket_mapping import BucketMapping
from skorecard.bucketers import UserInputBucketer


@pytest.fixture()
def example_features_bucket_map():
    """Generate example dict."""
    return {
        "LIMIT_BAL": BucketMapping(
            feature_name="LIMIT_BAL",
            type="numerical",
            map=[105000.0, 265000.0],
            right=True,
        ),
        "BILL_AMT1": BucketMapping(
            feature_name="BILL_AMT1",
            type="numerical",
            map=[34211.5, 173337.5],
            right=True,
        ),
    }


def test_no_transformation(df):
    """Test that we can use the transformer with no input."""
    X = df
    ui_bucketer = UserInputBucketer()

    new_X = ui_bucketer.transform(X)
    assert X.equals(new_X)


def test_manual_transformation(df, example_features_bucket_map):
    """Test that we can use an example dict for ManualBucketTransformer."""
    X = df
    y = df["default"]

    ui_bucketer = UserInputBucketer(example_features_bucket_map)
    new_X = ui_bucketer.fit_transform(X)
    assert len(new_X["LIMIT_BAL"].unique()) == 3
    assert len(new_X["BILL_AMT1"].unique()) == 3

    # specify y
    ui_bucketer = UserInputBucketer(example_features_bucket_map)
    ui_bucketer.fit(X, y)
    ui_bucketer.transform(X)


def test_bucket_table_method(df, example_features_bucket_map):
    """
    Make sure .bucket_table() works properly.
    """
    X = df
    y = df["default"]

    ui_bucketer = UserInputBucketer(example_features_bucket_map)

    with pytest.raises(NotFittedError):
        ui_bucketer.bucket_table("EDUCATION")

    ui_bucketer.fit(X, y)

    with pytest.raises(ValueError):
        # EDUCATION is not in 'example_features_bucket_map'
        ui_bucketer.bucket_table("EDUCATION")

    assert ui_bucketer.bucket_table("BILL_AMT1").shape[0] == 4


def test_summary_method(df, example_features_bucket_map):
    """
    Make sure .summary() works properly.
    """
    X = df
    y = df["default"]

    ui_bucketer = UserInputBucketer(example_features_bucket_map)

    with pytest.raises(NotFittedError):
        ui_bucketer.summary()

    ui_bucketer.fit(X, y)
    assert ui_bucketer.summary().shape[0] == 6
